import sys
from contextlib import contextmanager
from pathlib import Path
from unittest.mock import MagicMock

import pytest
import yaml
from pytest_mock import MockerFixture

from ampel.cli.main import main


@contextmanager
def argv_context(args: list[str]):
    argv = sys.argv
    try:
        sys.argv = args
        yield
    finally:
        sys.argv = argv


def run(args: list[str]) -> None | int | str:
    try:
        with argv_context(args):
            main()
        return None
    except SystemExit as se:
        return se.code


def dump(payload, tmpdir, name: str) -> Path:
    f = Path(tmpdir) / name
    with f.open("w") as fd:
        yaml.dump(payload, fd)
    return f


@pytest.fixture
def vault(tmpdir):
    return dump({"foo": "bar"}, tmpdir, "secrets.yml")


@pytest.fixture
def secrets():
    return {
        "str": "str",
        "key.with.bunch.of.dots": "nesty",
        "dict": {"user": "none", "password": 1.5},
        "list": [1, 2, 3.5, "flerp"],
    }


@pytest.fixture
def dir_secret_store(tmpdir, secrets):
    parent_dir = Path(tmpdir) / "secrets"
    parent_dir.mkdir()
    for k, v in secrets.items():
        dump(v, parent_dir, k)
    return parent_dir


@pytest.fixture
def schema(tmpdir):
    return dump({"name": "job", "task": [{"unit": "Nonesuch"}]}, tmpdir, "schema.yml")


@pytest.fixture
def mock_db(mocker: MockerFixture):
    return mocker.patch("ampel.core.AmpelDB.AmpelDB.get_collection")


@pytest.fixture
def mock_new_context_unit(mocker: MockerFixture, mock_db):
    return mocker.patch("ampel.core.UnitLoader.UnitLoader.new_context_unit")


def test_resource_passing(
    testing_config,
    mock_db: MagicMock,
    vault: Path,
    tmpdir,
):
    key = "dana"
    value = "zuul"

    writer = dump(
        {
            "unit": "DummyResourceOutputUnit",
            "config": {
                "name": key,
                "value": value,
            },
        },
        tmpdir,
        "writer.yml",
    )

    reader = dump(
        {
            "unit": "DummyResourceInputUnit",
            "config": {"value": key, "expected_value": value},
        },
        tmpdir,
        "reader.yml",
    )

    def run_task(task_path: Path, first: bool = False):
        assert (
            run(
                [
                    "ampel",
                    "process",
                    "--config",
                    str(testing_config),
                    "--secrets",
                    str(vault),
                    "--db",
                    "whatevs",
                    "--log-profile",
                    "console_debug",
                    "--schema",
                    str(task_path),
                    "--resources-in",
                    str(tmpdir / "resources.json") if not first else "",
                    "--resources-out",
                    str(tmpdir / "resources.json"),
                    "--name",
                    "task_1",
                ]
            )
            is None
        )

    run_task(writer, first=True)
    run_task(reader)


def test_templates(
    testing_config,
    mock_db: MagicMock,
    vault: Path,
    tmpdir,
):
    """ProcessCommand resolves templates"""
    task = dump(
        {
            "template": "hash_t2_config",
            "unit": "DummyIngestUnit",
            "config": dict(
                directives=[
                    dict(
                        channel="TEST",
                        ingest=dict(
                            combine=[
                                dict(
                                    unit="T1SimpleCombiner",
                                    state_t2=[
                                        dict(
                                            unit="DummyTiedStateT2Unit",
                                            config={
                                                "t2_dependency": [
                                                    {
                                                        "unit": "DummyStateT2Unit",
                                                        "config": {"foo": 37},
                                                    }
                                                ]
                                            },
                                        )
                                    ],
                                )
                            ]
                        ),
                    )
                ]
            ),
        },
        tmpdir,
        "task.yml",
    )

    def run_task(task_path: Path):
        assert (
            run(
                [
                    "ampel",
                    "process",
                    "--config",
                    str(testing_config),
                    "--secrets",
                    str(vault),
                    "--db",
                    "whatevs",
                    "--log-profile",
                    "console_debug",
                    "--schema",
                    str(task_path),
                    "--name",
                    "task_1",
                ]
            )
            is None
        )

    run_task(task)

    conf = mock_db("conf")
    # get the last config inserted
    doc = next(
        d.args[0]
        for d in reversed(conf.insert_one.call_args_list)
        if "unit" in d.args[0]
    )
    assert doc["unit"] == "DummyIngestUnit"
    config_id = doc["config"]["directives"][0]["ingest"]["combine"][0]["state_t2"][0][
        "config"
    ]

    def get_config(config_id: int):
        return next(
            (
                d.args[0]
                for d in reversed(conf.insert_one.call_args_list)
                if d.args[0].get("_id") == config_id
            ),
            None,
        )

    assert isinstance(config_id, int), "config was hashed"
    config = get_config(config_id)
    assert "t2_dependency" in config
    subconfig = get_config(config["t2_dependency"][0]["config"])
    subconfig.pop("_id")
    assert subconfig == {"foo": 37}, "config was hashed recursively"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/job/JobModel.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  05.04.2023
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, Literal

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.model.ChannelModel import ChannelModel
from ampel.model.job.EnvSpec import EnvSpec
from ampel.model.job.JobTaskModel import JobTaskModel
from ampel.model.job.MongoOpts import MongoOpts


class JobModel(AmpelBaseModel):

	name: None | str
	sig: None | int
	requirements: list[str] = []
	env: dict[str, EnvSpec] = {}
	channel: list[ChannelModel] = []
	alias: dict[Literal["t0", "t1", "t2", "t3", "t4"], dict[str, Any]] = {}
	mongo: MongoOpts = MongoOpts()
	task: list[JobTaskModel]


	def __init__(self, **kwargs):

		# Allow 'name' channel name alias to avoid 'channel[].channel'
		for c in kwargs.get('channel', []):
			if isinstance(c, dict) and 'name' in c:
				c["channel"] = c.pop("name")

		if isinstance(kwargs.get('task'), dict):
			kwargs['task'] = [kwargs['task']]

		super().__init__(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/builder/FirstPassConfig.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.10.2019
# Last Modified Date:  07.11.2025
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import json
from typing import Any, Literal

from ampel.config.builder.DisplayOptions import DisplayOptions
from ampel.config.collector.AliasConfigCollector import AliasConfigCollector
from ampel.config.collector.ChannelConfigCollector import ChannelConfigCollector
from ampel.config.collector.ConfigCollector import ConfigCollector
from ampel.config.collector.DBConfigCollector import DBConfigCollector
from ampel.config.collector.ForwardProcessConfigCollector import (
	ForwardProcessConfigCollector,
)
from ampel.config.collector.LoggingCollector import LoggingCollector
from ampel.config.collector.ProcessConfigCollector import ProcessConfigCollector
from ampel.config.collector.ResourceConfigCollector import ResourceConfigCollector
from ampel.config.collector.UnitConfigCollector import UnitConfigCollector
from ampel.log.AmpelLogger import AmpelLogger

tiers: tuple[Literal[0, 1, 2, 3], ...] = (0, 1, 2, 3)

class FirstPassConfig(dict):
	"""
	Class used to aggregate config pieces into a central configuration dict for ampel.
	"""

	conf_keys: dict[str, None | type[ConfigCollector]] = {
		"mongo": DBConfigCollector,
		"logging": LoggingCollector,
		"channel": ChannelConfigCollector,
		"unit": None,
		"process": None,
		"alias": None,
		"resource": ResourceConfigCollector,
	}

	def __init__(self,
		options: DisplayOptions,
		logger: None | AmpelLogger = None,
		ignore_exc: list[str] | None = None
	) -> None:

		self.logger = AmpelLogger.get_logger() if logger is None else logger

		d: dict[str, Any] = {
			k: Klass(conf_section=k, options=options, logger=logger)
			for k, Klass in self.conf_keys.items() if Klass
		}

		d['pwd'] = []
		d['unit'] = UnitConfigCollector(
			conf_section="unit", options=options, logger=logger, ignore_exc=ignore_exc
		)

		# Allow process to be defined in root key
		d['process'] = ForwardProcessConfigCollector(
			root_config=self, conf_section="process",
			target_collector_type=ProcessConfigCollector,
			logger=logger, options=options
		)

		d['alias'] = {}
		for k in tiers:

			d['alias'][f"t{k}"] = AliasConfigCollector(
				conf_section='alias', options=options, logger=logger, tier=k
			)

			# Allow processes to be defined in sub-tier entries already (process.t0, process.t1, ...)
			d['process'][f"t{k}"] = ProcessConfigCollector(
				conf_section='process', options=options, logger=logger, tier=k
			)

		d['process']["ops"] = ProcessConfigCollector(
			conf_section='process', options=options, logger=logger, tier="ops"
		)

		super().__init__(d)


	def unset_errors(self, d: None | dict = None) -> None:
		""" """
		for v in d.values() if d is not None else self.values():
			if isinstance(v, dict):
				if getattr(v, 'has_error', False):
					v.has_error = False # type: ignore[attr-defined]
				self.unset_errors(v)


	def has_nested_error(self, d=None, k=None) -> bool:

		ret = False

		for kk, dd in d.items() if d is not None else self.items():
			if isinstance(dd, dict):
				if getattr(dd, 'has_error', False):
					ret = True
					self.logger.warn(f"Errors occurred during configuration build in {dd.__class__.__name__}")
					if getattr(dd, 'err_fqns', False):
						self.logger.warn("[blue]───────[/blue]")
						self.logger.warn("[bold][blue]Summary[/bold][/blue]")
						self.logger.warn("[blue]───────\n[/blue]")
						for el in dd.err_fqns: # type: ignore[attr-defined]
							self.logger.warn(f"[red][bold]{el[0]}[/bold][/red]: {el[1]}")
						print(" ") # noqa
				if self.has_nested_error(dd, kk):
					ret = True

		return ret


	def print(self) -> None:

		if self.has_nested_error():
			self.logger.warn(
				"Warning: error were reported while collecting configurations"
			)

		self.logger.info(json.dumps(self, indent=4))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/log/__init__.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                11.06.2020
# Last Modified Date:  11.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

# flake8: noqa: F401
from .AmpelLogger import DEBUG, ERROR, INFO, SHOUT, VERBOSE, WARNING, AmpelLogger
from .LogFlag import LogFlag



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsUnitResultAdapter.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                20.04.2022
# Last Modified Date:  28.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.core.ContextUnit import ContextUnit
from ampel.struct.UnitResult import UnitResult
from ampel.types import Traceless


class AbsUnitResultAdapter(ContextUnit, AmpelABC, abstract=True):

	run_id: Traceless[int]

	@abstractmethod
	def handle(self, ur: UnitResult) -> UnitResult:
		raise NotImplementedError



from collections.abc import Iterable
from contextlib import contextmanager
from typing import Any

from typing_extensions import Self

from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.abstract.AbsIngester import AbsIngester
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.content.DataPoint import DataPoint
from ampel.content.StockDocument import StockDocument
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.model.UnitModel import UnitModel
from ampel.mongo.update.DBUpdatesBuffer import DBUpdatesBuffer
from ampel.mongo.update.MongoStockUpdater import MongoStockUpdater
from ampel.protocol.StockIngesterProtocol import StockIngesterProtocol
from ampel.types import OneOrMany, Tag


class _StockIngester:

    def __init__(self, ingester: AbsDocIngester[StockDocument], updater: MongoStockUpdater) -> None:
        self.ingester = ingester
        self.update = updater
    
    def ingest(self, doc: StockDocument) -> None:
        self.ingester.ingest(doc)

class _UpdatesBufferModel(AmpelBaseModel):
    max_size: int = 500
    push_interval: float = 3

class MongoIngester(AbsIngester):

    updates_buffer: _UpdatesBufferModel = _UpdatesBufferModel()

    #: Tag(s) to add to the stock :class:`~ampel.content.JournalRecord.JournalRecord`
    #: every time a document is processed
    jtag: None | OneOrMany[Tag]

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        updates_buffer = DBUpdatesBuffer(
            self.context.db,
            self.run_id,
            self.logger,
            error_callback=self.error_callback,
            acknowledge_callback=self.acknowledge_callback,
            max_size=self.updates_buffer.max_size,
            push_interval=self.updates_buffer.push_interval,
            raise_exc=self.raise_exc,
        )

        stock_updater = MongoStockUpdater(
            ampel_db = self.context.db, tier = self.tier, run_id = self.run_id,
            process_name = self.process_name, logger = self.logger,
            raise_exc = self.raise_exc, extra_tag = self.jtag

        )

        # Create ingesters
        dbconf = self.context.config.get("mongo.ingest", dict, raise_exc=True)

        def get_ingester_model(key: str) -> UnitModel:
            model = dbconf[key]
            if isinstance(model, str):
                return UnitModel(unit=model)
            return UnitModel(**model)

        self._t0 = AuxUnitRegister.new_unit(
            model=get_ingester_model("t0"),
            sub_type=AbsDocIngester[DataPoint],
            updates_buffer=updates_buffer,
        )

        self._t1 = AuxUnitRegister.new_unit(
            model=get_ingester_model("t1"),
            sub_type=AbsDocIngester[T1Document],
            updates_buffer=updates_buffer,
        )

        self._t2 = AuxUnitRegister.new_unit(
            model=get_ingester_model("t2"),
            sub_type=AbsDocIngester[T2Document],
            updates_buffer=updates_buffer,
        )

        self._stock = _StockIngester(
            AuxUnitRegister.new_unit(
                model=get_ingester_model("stock"),
                sub_type=AbsDocIngester[StockDocument],
                updates_buffer=updates_buffer,
            ),
            stock_updater,
        )

        self._updates_buffer = updates_buffer

    def __enter__(self) -> "Self":
        self._updates_buffer.__enter__()
        return super().__enter__()

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._updates_buffer.__exit__(exc_type, exc_value, traceback)
        self._stock.update.flush()

    @contextmanager
    def group(self, acknowledge_messages: None | Iterable[Any] = None):
        with self._updates_buffer.group_updates():
            yield
            for message in acknowledge_messages or []:
                self._updates_buffer.acknowledge_on_push(message)
            if len(self._stock.update._updates) >= self.updates_buffer.max_size:  # noqa: SLF001
                self._stock.update.flush()
    
    def flush(self) -> None:
        self._updates_buffer.push_updates(force=True)
        self._stock.update.flush()

    @property
    def stock(self) -> StockIngesterProtocol:
        return self._stock

    @property
    def t0(self) -> AbsDocIngester[DataPoint]:
        return self._t0

    @property
    def t1(self) -> AbsDocIngester[T1Document]:
        return self._t1

    @property
    def t2(self) -> AbsDocIngester[T2Document]:
        return self._t2



import sys
from contextlib import contextmanager, suppress
from pathlib import Path
from typing import Any

import pytest
import yaml
from pytest_mock import MockerFixture

from ampel.cli.main import main
from ampel.cli.utils import get_vault
from ampel.config.AmpelConfig import AmpelConfig
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.secret.AmpelVault import AmpelVault
from ampel.test.dummy import DummyInputUnit


@contextmanager
def argv_context(args: list[str]):
    argv = sys.argv
    try:
        sys.argv = args
        yield
    finally:
        sys.argv = argv


def run(args: list[Any]) -> None | int | str:
    try:
        with argv_context([str(e) for e in args]):
            main()
        return None
    except SystemExit as se:
        return se.code


def dump(payload, tmpdir, name: str) -> Path:
    f = Path(tmpdir) / name
    with f.open("w") as fd:
        yaml.dump(payload, fd)
    return f


@pytest.fixture
def vault(tmpdir):
    return dump({"foo": "bar"}, tmpdir, "secrets.yml")


@pytest.fixture
def secrets():
    return {
        "str": "str",
        "key.with.bunch.of.dots": "nesty",
        "dict": {"user": "none", "password": 1.5},
        "list": [1, 2, 3.5, "flerp"],
    }


@pytest.fixture
def dir_secret_store(tmpdir, secrets):
    parent_dir = Path(tmpdir) / "secrets"
    parent_dir.mkdir()
    for k, v in secrets.items():
        dump(v, parent_dir, k)
    return parent_dir


@pytest.fixture
def schema(tmpdir):
    return dump({"name": "job", "task": [{"unit": "Nonesuch"}]}, tmpdir, "schema.yml")


@pytest.fixture
def mock_db(mocker: MockerFixture):
    return mocker.patch("ampel.core.AmpelDB.AmpelDB.get_collection")


@pytest.fixture
def mock_new_context_unit(mocker: MockerFixture, mock_db):
    return mocker.patch("ampel.core.UnitLoader.UnitLoader.new_context_unit")


def test_secrets(testing_config, vault: Path, schema: Path, mock_new_context_unit):
    assert (
        run(
            [
                "ampel",
                "job",
                "--config",
                str(testing_config),
                "--no-conf-check",
                "--secrets",
                str(vault),
                "--schema",
                str(schema),
            ]
        )
        is None
    )
    assert mock_new_context_unit.call_count == 1
    loader_vault: None | AmpelVault = mock_new_context_unit.call_args.kwargs[
        "context"
    ].loader.vault
    assert loader_vault is not None
    secret = loader_vault.get_named_secret("foo")
    assert secret is not None
    assert secret.get() == "bar"


def test_resources_from_env(
    testing_config,
    vault: Path,
    schema: Path,
    mock_new_context_unit,
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.setenv("AMPEL_CONFIG_resource.mongo", "flerp")
    monkeypatch.setenv("AMPEL_CONFIG_resource.herp", "37")
    assert (
        run(
            [
                "ampel",
                "job",
                "--config",
                str(testing_config),
                "--no-conf-check",
                "--secrets",
                str(vault),
                "--schema",
                str(schema),
            ]
        )
        is None
    )
    config: AmpelConfig = mock_new_context_unit.call_args.kwargs["context"].config
    assert config.get("resource.mongo") == "flerp"
    assert config.get("resource.herp") == 37


def test_dir_secrets(dir_secret_store, secrets):
    vault = get_vault({"secrets": dir_secret_store})
    for k, v in secrets.items():
        assert vault.get_named_secret(k).get() == v

    assert vault.get_named_secret("nonesuch") is None, "missing secrets skipped cleanly"


def test_job_file(
    testing_config: Path,
    mock_context: DevAmpelContext,
    mocker: MockerFixture,
):
    proceed = mocker.patch.object(
        DummyInputUnit, "proceed", side_effect=DummyInputUnit.proceed, autospec=True
    )
    assert (
        run(
            [
                "ampel",
                "job",
                "--config",
                testing_config,
                "--no-conf-check",
                "--schema",
                testing_config.parent / "jobs" / "template_job.yaml",
            ]
        )
        is None
    )
    assert proceed.called


def test_job_db_prefix(
    testing_config: Path,
    integration_context,
):
    assert (
        run(
            [
                "ampel",
                "job",
                "--config",
                testing_config,
                "--no-conf-check",
                "--schema",
                testing_config.parent / "jobs" / "template_job2.yaml",
            ]
        )
        is None
    )
    db_names = integration_context.db.mongo_clients["writer"].list_database_names()
    right_custom_db_name = any(name == "_PyTestDB" for name in db_names)
	
    with suppress(Exception):
        integration_context.db.mongo_clients["writer"].drop_database("_PyTestDB")

    assert right_custom_db_name

    for client in integration_context.db.mongo_clients.values():
        client.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/log/handlers/AmpelStreamHandler.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                17.10.2018
# Last Modified Date:  06.11.2025
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import os, sys
from time import strftime
from typing import Literal

from rich.text import Text
from rich.console import Console

from ampel.log.LightLogRecord import LightLogRecord
from ampel.log.LogFlag import LogFlag
from ampel.util.collections import try_reduce
from ampel.util.mappings import compare_dict_values

levels = {1: 'DEBUG', 2: 'VERBOSE', 4: 'INFO', 8: 'SHOUT', 16: 'WARNING', 32: 'ERROR'}

class AmpelStreamHandler:
	"""
	:param int aggregate_interval: logs with similar attributes (log level, ...) are aggregated
	in one document instead of being split into several documents (spares some index RAM).
	*aggregate_interval* is the max interval of time in seconds during which log aggregation occurs.
	Beyond this value, a new log document is created no matter what.
	Note that it impacts the logging time granularity.

	:param density: whether log header (including timestamp) should be displayed on the console:

	- "default" (aggreation break if log-level, class-origin or 'extra' differ between log entries):
	2020-03-03 03:01:23 DistConfigBuilder VERBOSE
	 Loading conf/ampel-contrib-hu/unit.conf from distribution 'ampel-contrib-hu'

	2020-03-03 03:01:24 UnitConfigCollector VERBOSE
	 Adding T0 base unit: DecentFilter
	 Adding T0 base unit: SimpleDecentFilter
	 Adding T0 base unit: XShooterFilter

	- "compact" (aggregates log entries even if log entries originates from different classes):
	2020-03-03 03:01:23 DistConfigBuilder VERBOSE
	 Loading conf/ampel-contrib-hu/unit.conf from distribution 'ampel-contrib-hu'
	 Adding T0 base unit: DecentFilter
	 Adding T0 base unit: SimpleDecentFilter
	 Adding T0 base unit: XShooterFilter

	- "headerless" (Skip the log headers completely (no timestamp, no log level...):
	Loading conf/ampel-contrib-hu/unit.conf from distribution 'ampel-contrib-hu'
	Adding T0 base unit: DecentFilter
	Adding T0 base unit: SimpleDecentFilter
	Adding T0 base unit: XShooterFilter
	"""

	__slots__ = "__dict__", "print", "prev_record", "timefunc", \
		"prefix", "provenance", "nl", "aggregate_interval", \
		"datefmt", "timefmt", "cdate", "ctime", "cfile", "cline", "clvl"

	def __init__(self,
		std_stream: Literal['stdout', 'stderr'] = 'stderr',
		datefmt: str = "%Y-%m-%d ",
		timefmt: str = "%H:%M:%S ",
		level: int = LogFlag.INFO,
		aggregate_interval: float = 1.,
		density: Literal["default", "compact", "compacter", "headerless"] = "default",
		terminator: str = '\n',
		log_sep: str = '\n', # separator between aggregated log entries
		prefix: None | str = None,
		provenance: bool = True,
		color: bool = True,
		cdate: str = "cornflower_blue",    # rich color for date
		ctime: str = "light_sea_green",    # rich color for time
		# Note: dark_blue works too but salmon1 fits both light and dark backgrounds
		cfile: str = "salmon1",            # rich color for python file name
		cline: str = "rosy_brown",         # rich color for line number
		clvl: str = "grey53"               # rich color for log level
	) -> None:

		self.aggregate_interval = aggregate_interval
		self.dummy_record = LightLogRecord(name='dummy', levelno=1<<10, msg=None)
		self.prev_record: LightLogRecord = self.dummy_record
		if "%H" in datefmt:
			raise ValueError("Unexpected date format, please check/rebuild your config")
		self.datefmt = datefmt
		self.timefmt = timefmt
		self.prefix = prefix
		self.level = level
		self.nl = terminator
		self.log_sep = log_sep
		self.provenance = provenance
		self.ctime = ctime
		self.cdate = cdate
		self.cfile = cfile
		self.cline = cline
		self.clvl = clvl

		# Check env var NOCOLOR/NO_COLOR (case-insensitive, any non-empty value disables color)
		disable_color = (
			bool(os.environ.get("NOCOLOR", "").strip()) or
			bool(os.environ.get("NO_COLOR", "").strip())
		)

		# Disable colors when explicitly requested, when NOCOLOR is set or when output is redirected
		if not color or disable_color or not sys.stdout.isatty():
			self.console = Console(file=getattr(sys, std_stream), force_terminal=True, color_system=None)
		else:
			self.console = Console(file=getattr(sys, std_stream))

		self.print = self.console.print

		if density == "default":
			self.fields_check = ['extra', 'stock', 'channel', 'filename', 'levelno']
		elif density == "compact":
			self.fields_check = ['extra', 'stock', 'channel', 'levelno']
		elif density == "compacter":
			self.log_sep = ''
			self.fields_check = ['extra', 'stock', 'channel']
			self.handle = self.handle_compacter # type: ignore[method-assign]
		elif density == "headerless":
			self.handle = self.handle_headerless # type: ignore[method-assign]


	def handle(self, rec: LightLogRecord) -> None:

		prev_rec = self.prev_record

		# Same flag, date (+- 1 sec), tran_id and chans
		if not (
			rec.msg and rec.created - prev_rec.created < self.aggregate_interval and
			compare_dict_values(rec.__dict__, prev_rec.__dict__, self.fields_check)
		):
			self.print("", end=self.log_sep)
			self.print(self.format(rec))
			self.prev_record = rec

		if isinstance(rec.msg, str):
			self.print(" " + rec.msg.replace("\n", "\n "))


	def handle_compacter(self, rec: LightLogRecord) -> None:

		prev_rec = self.prev_record

		# Same flag, date (+- 1 sec), tran_id and chans
		if not (
			rec.msg and rec.created - prev_rec.created < self.aggregate_interval and
			compare_dict_values(rec.__dict__, prev_rec.__dict__, self.fields_check)
		):
			self.print("", end=self.log_sep)
			self.print(self.format(rec))
			self.prev_record = rec

		if isinstance(rec.msg, str):
			self.print(" " + rec.msg.replace("\n", "\n "))


	def handle_headerless(self, rec: LightLogRecord) -> None:
		if rec.msg:
			self.print(rec.msg)
		if rec.extra:
			if rec.msg:
				self.print(" ")
			self.print(str(rec.extra))


	def format(self, record: LightLogRecord) -> Text:

		t = Text()
		t.append(strftime(self.datefmt), style=self.cdate)
		t.append(strftime(self.timefmt), style=self.ctime)

		if self.prefix:
			t.append(self.prefix + " ")

		# Location (deactivated, takes space, no real benefit)
		"""
		lvl = record.levelno
		if lvl & 128:
			t.append(' UNIT')
		if lvl & 256:
			t.append(' CORE')
		"""

		if self.provenance and record.filename:
			if record.filename[0] == '<': # ipython
				t.append(record.filename, style=self.cfile)
			else:
				t.append(record.filename[:-3], style=self.cfile)
				t.append(":")
				t.append(str(record.lineno), style=self.cline)

		t.append(" " + levels[record.levelno >> 9], style=self.clvl)

		if record.extra:
			suffix = [f'{k}={record.extra[k]}' for k in record.extra]
		else:
			suffix = []

		if record.stock:
			suffix.append(f's={record.stock}') # type: ignore[str-bytes-safe]

		if record.channel:
			suffix.append(f'c={try_reduce(record.channel)}')

		if getattr(record, 'unit', None):
			suffix.append(f'unit={record.unit}')

		if suffix:
			t.append(f" [{', '.join(suffix)}]")

		return t


	def flush(self) -> None:
		self.console.file.flush()


	def break_aggregation(self) -> None:
		self.prev_record = self.dummy_record



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/AbsStockCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                25.03.2021
# Last Modified Date:  25.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, Literal

from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.LoadJSONAction import LoadJSONAction
from ampel.cli.MaybeIntAction import MaybeIntAction
from ampel.model.time.TimeConstraintModel import TimeConstraintModel
from ampel.model.time.TimeDeltaModel import TimeDeltaModel
from ampel.model.time.TimeLastRunModel import TimeLastRunModel
from ampel.model.time.TimeStringModel import TimeStringModel
from ampel.model.time.UnixTimeModel import UnixTimeModel
from ampel.model.UnitModel import UnitModel
from ampel.mongo.utils import maybe_match_array


class AbsStockCommand(AbsCoreCommand, abstract=True):
	"""
	Base class for commands selecting/matching stock(s)
	"""

	@staticmethod
	def get_select_args_help() -> dict[str, str]:

		return {
			# Required
			'config': 'Path to an ampel config file (yaml/json)',
			# Optional
			'secrets': 'Path to a YAML secrets store in sops format',
			'log-profile': 'One of: default, compact, headerless, verbose, debug',
			'id-mapper': 'Convert stock ids using the provided id mapper (ex: ZTFIdMapper)',
			# Selection
			'stock': 'Stock id(s) (OR matched if multi-valued)',
			'channel': 'Channel(s)',
			'created-after-ts': 'Created after unix timestamp',
			'created-after-str': 'Created after date-time iso string',
			'created-after-delta': 'Created after time delta',
			'created-after-process': 'Created after last run of process with name',
			'created-before-ts': 'Created before unix timestamp',
			'created-before-str': 'Created before date-time iso string',
			'created-before-delta': 'Created before time delta',
			'created-before-process': 'Created before last run of process with name',
			'updated-after-ts': 'Updated after unix timestamp',
			'updated-after-str': 'Updated after date-time iso string',
			'updated-after-delta': 'Updated after time delta',
			'updated-after-process': 'Updated after last run of process with name',
			'updated-before-ts': 'Updated before unix timestamp',
			'updated-before-str': 'Updated before date-time iso string',
			'updated-before-delta': 'Updated before time delta',
			'updated-before-process': 'Updated before last run of process with name',
			'custom-match': 'Custom mongodb match as JSON string (ex: {"body.aKey": {"$gt": 1}})',
		}


	def add_selection_args(self, builder: ArgParserBuilder, sub_ops: str = 'all') -> None:

		# Selection args
		builder.add_group('match', 'Stock selection arguments', sub_ops)
		builder.arg('stock', group='match', sub_ops=sub_ops, action=MaybeIntAction, nargs='+')
		builder.xargs(
			group = 'match',
			sub_ops = sub_ops,
			xargs = [
				{'name': 'created-before-str'}, {'name': 'created-before-ts', 'type': int},
				{'name': 'created-before-delta', 'action': LoadJSONAction},
				{'name': 'created-before-process'}
			]
		)
		builder.xargs(
			group = 'match',
			sub_ops = sub_ops,
			xargs = [
				{'name': 'created-after-str'}, {'name': 'created-after-ts', 'type': int},
				{'name': 'created-after-delta', 'action': LoadJSONAction},
				{'name': 'created-after-process'}
			]
		)
		builder.xargs(
			group = 'match',
			sub_ops = sub_ops,
			xargs = [
				{'name': 'updated-before-str'}, {'name': 'updated-before-ts', 'type': int},
				{'name': 'updated-before-delta', 'action': LoadJSONAction},
				{'name': 'updated-before-process'}
			]
		)
		builder.xargs(
			group = 'match',
			sub_ops = sub_ops,
			xargs = [
				{'name': 'updated-after-str'}, {'name': 'updated-after-ts', 'type': int},
				{'name': 'updated-after-delta', 'action': LoadJSONAction},
				{'name': 'updated-after-process'}
			]
		)
		builder.logic_args('channel', descr='Channel', group='match', sub_ops=sub_ops)
		builder.logic_args('with-tag', descr='Tag', group='match', sub_ops=sub_ops)
		builder.logic_args('without-tag', descr='Tag', group='match', sub_ops=sub_ops, excl=True)
		builder.arg(
			'custom-match', group='match', sub_ops=sub_ops,
			metavar='#', action=LoadJSONAction
		)


	def get_tag(self, args: dict[str, Any]) -> None | dict[Literal['with', 'without'], dict]:

		tag: None | dict[Literal['with', 'without'], dict] = None
		if args.get('with_tag'):
			tag = {'with': args['with_tag']}
		if args.get('without_tag'):
			if tag is None:
				tag = {}
			tag['without'] = args['without_tag']
		return tag


	def build_select_model(self, args: dict[str, Any]) -> UnitModel:

		conf = {
			'created': self.get_time_model('created', args),
			'updated': self.get_time_model('updated', args),
			'channel': args['channel'],
			'custom': args['custom_match']
		}

		if args.get('tag'):
			conf['tag'] = self.get_tag(args)

		if (stock := args.get('stock')):
			conf['custom'] = { # Hmmm, unsure about the necessity of 'custom' there, check later
				'stock': stock if isinstance(stock, int | bytes | str)
					else maybe_match_array(stock)
			}

		return UnitModel(unit='T3StockSelector', config=conf)


	def get_time_model(self, prefix: str, args: dict[str, Any]) -> TimeConstraintModel:

		d: dict[str, Any] = {'after': None, 'before': None}

		for when in ('after', 'before'):
			if args.get(x := f'{prefix}_{when}_ts'):
				d[when] = UnixTimeModel(match_type='unix_time', value=args[x])
			elif args.get(x := f'{prefix}_{when}_str'):
				d[when] = TimeStringModel(
					match_type='time_string',
					dateTimeStr=args[x],
					dateTimeFormat='%Y%m%dT%H%M%S'
				)
			elif args.get(x := f'{prefix}_{when}_delta'):
				d[when] = TimeDeltaModel(match_type='time_delta', **args[x])
			elif args.get(x := f'{prefix}_{when}_process'):
				d[when] = TimeLastRunModel(match_type='time_last_run', process_name=args[x])

		return TimeConstraintModel(**d)



from ampel.util.mappings import unflatten_dict


def test_unflatten_dict():
    assert unflatten_dict({"count.chans.HU_SN": 10}) == {
        "count": {"chans": {"HU_SN": 10}}
    }

    assert unflatten_dict(
        {
            "_id": 1,
            "tag": 1,
            "excl": 1,
            "body.jd": 1,
            "body.fid": 1,
            "body.rcid": 1,
            "body.magpsf": 1,
        }
    ) == {
        "_id": 1,
        "tag": 1,
        "excl": 1,
        "body": {"jd": 1, "fid": 1, "rcid": 1, "magpsf": 1},
    }

    assert unflatten_dict(
        {"a.0.b.f.0": 1, "a.0.b.f.1": 2, "a.0.b.f.2": 3, "a.1.c": 2, "d.e": 1},
        unflatten_list=True,
    ) == {"a": [{"b": {"f": [1, 2, 3]}}, {"c": 2}], "d": {"e": 1}}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/query/var/LogsLoader.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                29.11.2018
# Last Modified Date:  16.03.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Sequence
from typing import Any, Literal

from pymongo.collection import Collection

from ampel.base.AmpelFlexModel import AmpelFlexModel
from ampel.content.LogDocument import LogDocument
from ampel.log.LogFlag import LogFlag
from ampel.view.ReadOnlyDict import ReadOnlyDict


class LogsLoader(AmpelFlexModel):
	"""
	Abbreviations:
	s: stock, a: alert, f: flag, r: run, m: msg, c: channel

	:param decompactify:
	:param resolve_flag: load flag int as LogFlag
		(repr becomes <LogFlag.SCHEDULED_RUN|CORE|T2|INFO: 8836> instead of 8836)
	:param bool simplify:
	:param datetime_ouput: if not None, datetime of each log entry (extracted from ObjectId)
		will be added to the log content (use param datetime_key to configure whished key).
		Possible values:
		- 'string': datetime will be added as string: '2019-03-07T08:32:17.000Z'
		- 'date': datetime will be added as datetime object: datetime.datetime(2019, 3, 7, 8, 32, 30)
	:param datetime_key: '_id': garanteed in first position but overrides default _id (don't use remove=["_id"]...)
		or any string ('date' for example)
	:param debug: print generated aggregation stages if True
	:param remove_keys: remove key (possible values: 'c', '_id', 's') during projection stage
	"""

	decompactify: bool = True
	simplify: bool = False
	hexify: bool = True
	resolve_flag: bool = True
	read_only: bool = False
	remove_keys: None | list[str] = ["channel"]
	datetime_ouput: Literal['string', 'date'] = 'string'
	datetime_key: str = '_id'
	verbose: bool = False
	debug: bool = False


	def fetch_logs(self,
		col: Collection,
		match: None | dict[str, Any] = None,
		channel: None | dict[str, Any] = None
	) -> Sequence[LogDocument]:
		"""
		:param match: match criteria
		:param channel: The $unwind aggregagtion pipeline stage requires to re-apply channel match criteria.
			These are not necessarily fetchable using match.get('channel') since complex match logic can use the $or operator.
			(No need to use this argument if match['channel'] exists, it is used automatically in this case)
		"""

		stages = [
			# Matching criteria (can contain nested dicts in case of complex criteria)
			{'$match': match or {}}
		]

		# Extract datetime from objectid and add it as 'date' field
		if self.datetime_ouput:
			stages.append(
				{
					'$addFields': {
						self.datetime_key: {
							"$convert": {
								'input': {'$toDate': '$_id'},
								'to': self.datetime_ouput
							}
						}
					}
				}
			)

		# Unwind, i.e converts
		#  {'_id': ObjectId('5c80d71154048002ca372208'),
		#  'f': 8740,
		#  'r': 714,
		#  'c': 3,
		#  'm': ['msg1', 'msg2'],
		# Into:
		#  {'_id': ObjectId('5c80d71154048002ca372208'),
		#  'f': 8740,
		#  'r': 714,
		#  'c': 3,
		#  'm': 'msg1'},
		#  {'_id': ObjectId('5c80d71154048002ca372208'),
		#  'f': 8740,
		#  'r': 714,
		#  'c': 3,
		#  'm': 'msg2'},
		stages.append(
			{
				'$unwind': {
					'path': '$m',
					'preserveNullAndEmptyArrays': True
				}
			}
		)

		if self.decompactify:

			# Converts
			#  {'_id': ObjectId('5c80d71154048002ca372208'),
			#  's': 1810112413252531,
			#  'a': 697252385515010002,
			#  'f': 8740,
			#  'r': 714,
			#  'm': {'c': 3, 't': 'test2'},
			#  'date': datetime.datetime(2019, 3, 7, 8, 32, 17)}
			# Into:
			#  {'_id': ObjectId('5c80d71154048002ca372208'),
			#  's': 1810112413252531,
			#  'a': 697252385515010002,
			#  'f': 8740,
			#  'r': 714,
			#  'm': 'test2',
			#  'date': datetime.datetime(2019, 3, 7, 8, 32, 17),
			#  'c': 3}
			#######################################################

			# Part 1
			# a) Copy channel info embedded in compact msg to log entry root level
			# b) If m.m exists, copy it to root level as 'msg'
			stages.append(
				{
					"$addFields": {
						"c": {
							"$ifNull": ["$c", "$m.c"]
						},
						"m": {
							"$ifNull": [
								"$m.m",
								{"$ifNull": ["$m", "$REMOVE"]}
							]
						},
					}
				}
			)

			# Part 2:
			# remove compact dict leftover
			# ('msg': {'channel': 3, 'txt': 'test2'})
			stages.append(
				{
					"$addFields": {
						"m": {
							"$cond": {
								"if": {"$not": "$m.c"},
								"then": "$m",
								"else": "$REMOVE"
							}
						}
					}
				}
			)

			# re-apply channel filter as $unwind may have included
			# log entries associated with other channel
			if (match and (c := match.get('channel'))) or channel:
				stages.append(
					{'$match': {'c': c or channel}}
				)

		if self.remove_keys:

			# Projection (last aggregation stage)
			proj: dict[str, Any] = {'$project': {}}

			for k in self.remove_keys:
				proj['$project'][k] = 0

			stages.append(proj)

		if self.debug:
			print(f"Using aggregation: {stages}")  # noqa: T201

		log_entries: list[LogDocument] = list(col.aggregate(stages))

		if self.resolve_flag:
			for el in log_entries:
				el["f"] = LogFlag(el["f"])

		if self.simplify:
			for el in log_entries:
				print(f'{el["_id"]!r} {el["m"]}')  # noqa: T201
			return []

		# if hexify:
		#	for el in log_entries:
		#		if 'e' in el:
		#			el['e']['comp'] = el['e']['comp'].hex()

		if self.read_only:
			return tuple(ReadOnlyDict(el) for el in log_entries) # type: ignore[misc]

		return log_entries



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/update/DBUpdatesBuffer.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                31.10.2018
# Last Modified Date:  22.04.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Callable, Generator, Iterable, Iterator, Mapping
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from threading import Event, Lock
from typing import Any, Literal

from pymongo import InsertOne, UpdateMany, UpdateOne
from pymongo.collection import Collection
from pymongo.errors import BulkWriteError

from ampel.core.AmpelDB import AmpelDB, intcol
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.utils import convert_dollars, report_error, report_exception
from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry

DBOp = UpdateOne | UpdateMany | InsertOne
AmpelMainCol = Literal['stock', 't0', 't1', 't2', 't3']

# Monitoring counters
stat_db_ops = AmpelMetricsRegistry.counter(
	"ops",
	"Number of bulk operations",
	subsystem="db",
	labelnames=("col",)
)
stat_db_errors = AmpelMetricsRegistry.counter(
	"errors",
	"Number of bulk op errors",
	subsystem="db",
	labelnames=("col",)
)
stat_db_time = AmpelMetricsRegistry.histogram(
	"time",
	"Latency of bulk operations",
	unit="seconds",
	subsystem="db",
	labelnames=("col",)
)

class DBUpdatesBuffer:
	"""
	TODO:
	* Try to mark transient docs on error
	* Do something with self._err_db_ops

	Note1:
	Regarding multithreading: pymongo uses the standard Python socket module,
	which can drop the GIL while sending and receiving data over the network.

	Note2:
	On the advantages of buffering updates and submitting them in bulk

	-> Without buffering:
	In []:
		start=time()
		for i in range(200000):
			col.insert_one({'_id':i})
		print(f"{round(time()-start, 1)}s")
	Out []: 60.4s


	-> With buffering:
	In []:
		ops=[]
		start=time()
		for i in range(200000):
			ops.append(InsertOne({'_id': i}))
			if i % 2000 == 0:
				col.bulk_write(list(ops), ordered=False)
				ops=[]
		col.bulk_write(ops, ordered=False)
		print(f"{round(time()-start, 1)}s")
	Out []: 1.7s
	"""

	def __init__(self,
		ampel_db: AmpelDB,
		run_id: int | list[int],
		logger: AmpelLogger,
		error_callback: None | Callable[[], None] = None,
		acknowledge_callback: None | Callable[[Iterator[Any]], None] = None,
		log_doc_ids: None | Iterable[int] = None,
		push_interval: None | float = 3,
		max_size: None | int = None,
		raise_exc: bool = False
	):
		"""
		:param error_callback: callback method to be called on errors
		:param acknowledge_callback: method to call to acknowledge processing of a message bunch
		:param catch_signals: see Schedulable docstring
		:param log_doc_ids: logs inserted/updated document IDs for the given collections.
		Collection integer identifiers are: 't0' -> 0, 't1' -> 1, 't2' -> 2, 'stock' -> 3.
		Thus, if you want to activate this behavior for all collections, use log_doc_ids=[0, 1, 2, 3].
		Notes: 1) it this will significantly increase the log size/amount.
		2) the "_id" of T2 documents is generated server-side and bulk_write does not return it,
		so it is not avail 'as is'. We instead log the natching filter (t2id, config, link)
		which is guaranteed to be unique, so it's an equivalent.
		3) Huge values of push_interval can yield a document size overflow (16MB limit per doc)
		:param push_interval: If provided, updates will be flushed every x seconds.
		:param max_size: If provided, updates will be flushed as soon as possible after the
		number of updated in any collection (stock, t0, t1, t2) exceeds the provided limit.
		:param threads: perform the various collection bound bulk_write() operations in dedicated threads.
		The provided integer number defines the size of the thread pool. Note that no real performance gain
		was yet noticed using a meaningful value such as 4 (OSX, python 3.8.1). Since bulk_write drops the GIL,
		a multithreading-like effect already occurs without the specific use a threads.
		"""

		self._new_buffer()
		self.error_callback = error_callback
		self.acknowledge_callback = acknowledge_callback

		self._cols: dict[AmpelMainCol, Collection] = {
			col_name: ampel_db.get_collection(col_name)
			for col_name in self.db_ops
		}

		self.stats: dict[AmpelMainCol, int] = {
			'stock': 0, 't0': 0, 't1': 0, 't2': 0,
		}

		self._err_db_ops: dict[str, Any] = {k: [] for k in self.db_ops}
		self._ampel_db = ampel_db
		self.run_id = run_id
		self.logger = logger
		self.max_size = max_size
		self.log_doc_ids = set(log_doc_ids) if log_doc_ids else None

		# wake up update pusher thread
		self._push = Event()
		# prevent autopush while we are in the group_updates context
		self._block_autopush = Lock()
		self._pushing = Lock()
		# signal update pusher thread to stop
		self._stop = Event()
		self._exec = ThreadPoolExecutor()

		self.push_interval = push_interval
		self.raise_exc = raise_exc


	def _new_buffer(self) -> None:
		""" Creates new buffer for pymongo operations """

		# total number of updates (of all kinds/collections)
		self.db_ops: dict[AmpelMainCol, list[DBOp]] = {
			't2': [], 't0': [], 'stock': [], 't1': []
		}
		self._messages_to_ack: list[Any] = []

	def __enter__(self) -> None:
		"""
		:raises: RuntimeError if context manager is already running
		"""
		self._stop.clear()
		def task():
			while not self._stop.is_set():
				self._push.wait(self.push_interval)
				self._push.clear()
				self.push_updates(force=True)
		self._task = self._exec.submit(task)
	
	def __exit__(self, exc_type: type[BaseException], exc_value: BaseException, traceback: Any) -> None:
		"""
		:raises: RuntimeError if context manager is already running
		"""
		# stop updater thread
		self._stop.set()
		self._push.set()
		self._task.result()
		# push any remaining updates
		self.push_updates()


	def acknowledge_on_push(self, message: Any) -> None:
		"""
		Add a message to be passed to acknowledge_callback when the current
		buffer has been pushed
		"""
		if self.acknowledge_callback:
			self._messages_to_ack.append(message)


	def add_updates(self, updates: dict[AmpelMainCol, list[DBOp]]) -> None:
		"""
		:raises: KeyError if dict key is unknown (known keys: stock, t0, t1, t2)
		"""
		for k, v in updates.items():
			self.db_ops[k] += v


	def add_col_updates(self, col: AmpelMainCol, updates: list[DBOp]) -> None:
		""" :raises: KeyError if col is unknown (known cols: stock, t0, t1, t2) """
		self.db_ops[col] += updates


	def add_col_update(self, col: AmpelMainCol, update: DBOp) -> None:
		""" :raises: KeyError if col is unknown (known cols: stock, t0, t1, t2) """
		self.db_ops[col].append(update)


	def add_t0_update(self, update: DBOp) -> None:
		self.db_ops['t0'].append(update)


	def add_t1_update(self, update: DBOp) -> None:
		self.db_ops['t1'].append(update)


	def add_t2_update(self, update: DBOp) -> None:
		self.db_ops['t2'].append(update)


	def add_stock_update(self, update: DBOp) -> None:
		self.db_ops['stock'].append(update)


	@contextmanager
	def group_updates(self) -> Generator:
		"""
		Ensure that updates issued in this context are grouped together
		"""
		with self._block_autopush:
			yield
			self.check_push()
		# potentially raise exceptions to main thread
		if self._task.done():
			self._task.result()


	def check_push(self) -> None:
		if self.max_size is not None and any(
			len(v) > self.max_size
			for v in self.db_ops.values()
		):
			# block until the previous buffer is pushed, then signal the thread
			# to swap and push the new one
			with self._pushing:
				self._push.set()


	def push_updates(self, force: bool = False) -> None:

		# swap buffers
		with self._block_autopush:
			db_ops, messages = self.db_ops, self._messages_to_ack
			self._new_buffer()

		# prevent the new buffer from overfilling before bulk writes complete
		with self._pushing:
			for _ in self._exec.map(
				self.call_bulk_write,
				*zip(*((col_name, ops) for col_name, ops in db_ops.items() if ops), strict=True)
			):
				pass

			if self.acknowledge_callback and messages:
				try:
					self.acknowledge_callback(iter(messages))
				except Exception as exc:
					if self.raise_exc:
						raise
					report_exception(self._ampel_db, self.logger, exc=exc)


	def call_bulk_write(self, col_name: AmpelMainCol, db_ops: list, *, extra: None | dict = None) -> None:
		"""
		:param col_name: Ampel DB collection name (ex: stock, t0, t1, t2)
		:param db_ops: list of pymongo operations
		:raises: None, but stops the AlertConsumer processing by using the method
		cancel_run() when unrecoverable exceptions occur.

		Regarding the handling of BulkWriteError:
		Concurent upserts triggers a DuplicateKeyError exception.

		https://stackoverflow.com/questions/37295648/mongoose-duplicate-key-error-with-upsert
		<quote>
			An upsert that results in a document insert is not a fully atomic operation.
			Think of the upsert as performing the following discrete steps:
				Query for the identified document to upsert.
				If the document exists, atomically update the existing document.
				Else (the document doesn't exist), atomically insert a new document
				that incorporates the query fields and the update.
		</quote>

		There are many tickets opened on the mongoDB bug tracker regarding this issue.
		One of which: https://jira.mongodb.org/browse/SERVER-14322
		where is stated:
			"It is expected that the client will take appropriate action
			upon detection of such constraint violation"

		All in all: the server behaves inappropriately, the driver won't catch those
		cases for us, so we have to do the work by ourself.

		Last: the use of SON (serialized Ocument Normalisation) is deprecated according
		to the mongoDB doc. It will be removed with pymongo 4, so we should not use it anymore.
		BUT: the offending updates (UpdateOne instances) returned by the server are
		provided as SON by BulkWriteError (array 'writeErrors' contains SON objects).
		So we have no other choice than handling with them for now.
		"""
		
		with stat_db_time.labels(col_name).time():
			try:

				# Update DB
				db_res = self._cols[col_name].bulk_write(db_ops, ordered=False)
				stat_db_ops.labels(col_name).inc(len(db_ops))

				self.logger.debug(
					None, extra=self._build_log_extra(
						col_name, db_ops, db_res.bulk_api_result, extra
					)
				)

				return

			except BulkWriteError as bwe:

				try:

					dup_key_only = True

					for err_dict in bwe.details.get('writeErrors', []):

						stat_db_errors.labels(col_name).inc()
						# 'code': 11000, 'errmsg': 'E11000 duplicate key error collection: ...
						if err_dict.get("code") == 11000:

							self.logger.info(
								f"Race condition during ingestion in '{col_name}': {err_dict}"
							)
							self.logger.flush()

							# Should no longer raise pymongo.errors.DuplicateKeyError
							self._cols[col_name].update_one(
								err_dict['op']['q'],
								err_dict['op']['u'],
								upsert=err_dict['op']['upsert']
							)
							stat_db_ops.labels(col_name).inc()

						else:

							dup_key_only = False
							self._err_db_ops[col_name].append(err_dict)

							# Try to insert doc into trouble collection (raises no exception)
							# Possible exception will be logged out to console in any case
							report_error(
								self._ampel_db, msg="BulkWriteError entry details",
								logger=self.logger, info={
									'run': self.run_id,
									'err': convert_dollars(err_dict)
								}
							)

							################################
							# TODO: better than this.
							# - Mark corresponding transients with an error flag (add channel info?)
							# - Implement something for temp DB connectivity issues ?
							################################

							if self.raise_exc:
								raise

					if dup_key_only:
						self.logger.debug(
							f"Race condition(s) recovered: {len(bwe.details.get('writeErrors', []))}",
							extra=self._build_log_extra(col_name, db_ops, bwe.details, extra)
						)

						return

				except Exception as ee:
					if self.raise_exc:
						raise
					# Log exc and try to insert doc into trouble collection (raises no exception)
					report_exception(self._ampel_db, self.logger, exc=ee)

			except Exception as e:
				if self.raise_exc:
					raise
				# Log exc and try to insert doc into trouble collection (raises no exception)
				report_exception(self._ampel_db, self.logger, exc=e)

			print(f"Update of {col_name} collection has failed")  # noqa: T201
			print(db_ops)  # noqa: T201

			self._err_db_ops[col_name] += db_ops
			if self.error_callback:
				self.error_callback()


	def _build_log_extra(self,
		col_name: str,
		ops: list[DBOp],
		bulk_api_result: Mapping[str, Any],
		extra: None | dict[str, Any] = None
	) -> dict[str, Any]:

		ret = {
			'col': intcol[col_name],
			'ins': bulk_api_result['nInserted'],
			'ups': bulk_api_result['nUpserted'],
			'mod': bulk_api_result['nModified']
		}

		if self.log_doc_ids and ret['col'] in self.log_doc_ids:
			try:
				if ret['col'] != 2:
					ret['docs'] = [op._filter['_id'] for op in ops] # type: ignore[union-attr]  # noqa: SLF001
				else:
					ret['docs'] = [
						{
							'unit': op._filter['unit'], # type: ignore[union-attr]  # noqa: SLF001
							'config': op._filter['config'], # type: ignore[union-attr]  # noqa: SLF001
							'link': op._filter['link'] # type: ignore[union-attr]  # noqa: SLF001
						}
						for op in ops
					]
			except Exception:
				pass

		if extra:
			return {**extra, **ret}

		return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/update/MongoT0Ingester.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                23.05.2021
# Last Modified Date:  14.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections import defaultdict
from datetime import datetime
from typing import Any, Literal

from pymongo import UpdateMany, UpdateOne

from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.content.DataPoint import DataPoint
from ampel.mongo.update.HasUpdatesBuffer import HasUpdatesBuffer
from ampel.mongo.utils import maybe_use_each
from ampel.types import DataPointId


class MongoT0Ingester(AbsDocIngester[DataPoint], HasUpdatesBuffer):
	""" Inserts `DataPoint` into the t0 collection  """

	#: If 1 or 2: raise DuplicateKeyError on attempts to upsert the same id with different bodies
	#: 1: partial check is performed (compatible with potential muxer projection)
	#: 2: strict check is performed
	extended_match: Literal[0, 1, 2] = 0
	update_ttl: bool = False

	def ingest(self, doc: DataPoint) -> None:

		match: dict[str, Any] = {'id': doc['id']}
		upd: dict[str, Any] = {
			'$addToSet': {
				'channel': maybe_use_each(doc['channel'])
			},
			'$push': {
				'meta': maybe_use_each(doc['meta']) # meta must be set by compiler
			}
		}

		if 'body' in doc:
			upd['$setOnInsert'] = {'body': doc['body']}
			if self.extended_match:
				if self.extended_match == 2:
					match['body'] = doc['body']
				else:
					for k, v in doc['body'].items():
						match['body.' + k] = v

		if 'origin' in doc:
			match['origin'] = doc['origin']

		if 'tag' in doc:
			upd['$addToSet']['tag'] = maybe_use_each(doc['tag'])

		if 'stock' in doc:
			upd['$addToSet']['stock'] = doc['stock'] if isinstance(doc['stock'], int | bytes | str) \
				else maybe_use_each(doc['stock'])

		if 'expiry' in doc:
			upd['$max'] = {'expiry': doc['expiry']}

		self.updates_buffer.add_t0_update(
			UpdateOne(match, upd, upsert=True)
		)

	def update_expiry(self, dps: dict[DataPointId, datetime]) -> None:
		by_ttl: dict[datetime, list[DataPointId]] = defaultdict(list)
		for k, v in dps.items():
			by_ttl[v].append(k)
		for expires, ids in by_ttl.items():
			self.updates_buffer.add_t0_update(
				UpdateMany(
					{'id': {'$in': ids}},
					{'$max': {'expiry': expires}}
				)
			)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/mongo/model/ShortIndexModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.04.2020
# Last Modified Date:  30.10.2025
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, Dict # noqa: UP035

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.mongo.model.FieldModel import FieldModel


class ShortIndexModel(AmpelBaseModel):

	field: str
	args: None | Dict[str, Any] = None # noqa: UP006

	def dict(self, **kwargs) ->  Dict[str, Any]: # noqa: UP006
		if self.args:
			return {
				"index": [(self.field, 1)],
				"args": self.args
			}
		return {"index": [(self.field, 1)]}


	def get_id(self) -> str:
		"""
		Returns an indexId similar to what pymongo index_information outputs.
		Ex: [('tranId', 1), ('channel', 1)] -> tranId_1_channel_1
		"""
		return FieldModel(field=self.field).get_id()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsWorker.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                28.05.2021
# Last Modified Date:  09.11.2025
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import gc
import random
import signal
from collections.abc import Generator, Sequence
from contextlib import contextmanager
from math import ceil
from threading import Event
from time import time
from typing import Any, ClassVar, Generic, Literal, TypeVar

from pymongo.read_concern import ReadConcern
from pymongo.write_concern import WriteConcern

from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.abstract.AbsIngester import AbsIngester
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.base.decorator import abstractmethod
from ampel.base.LogicalUnit import LogicalUnit
from ampel.content.MetaRecord import MetaRecord
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.core.EventHandler import EventHandler
from ampel.enum.DocumentCode import DocumentCode
from ampel.enum.EventCode import EventCode
from ampel.enum.MetaActionCode import MetaActionCode
from ampel.log import DEBUG, VERBOSE, AmpelLogger, LogFlag
from ampel.log.utils import convert_dollars, report_error, report_exception
from ampel.metrics.AmpelMetricsRegistry import (
	AmpelMetricsRegistry,
	Histogram,
	Summary,
	TimingCounter,
)
from ampel.model.UnitModel import UnitModel
from ampel.mongo.update.MongoIngester import MongoIngester
from ampel.mongo.utils import maybe_match_array, maybe_use_each
from ampel.types import JDict, OneOrMany, Tag, UBson
from ampel.util.hash import build_unsafe_dict_id
from ampel.util.tag import merge_tags

T = TypeVar("T", T1Document, T2Document)

class BackoffConfig(AmpelBaseModel):
	base: float = 2
	factor: float = 1
	jitter: bool = True
	max_value: float = 10.
	max_tries: None | int = None
	max_time: None | float = 60.

stat_time = AmpelMetricsRegistry.summary(
    "time",
    "Processing time",
    unit="microseconds",
    subsystem="worker",
    labelnames=("tier", "phase", "unit"),
)

@contextmanager
def stop_on_signal(signals: Sequence[signal.Signals], logger: AmpelLogger) -> Generator[Event, None, None]:
	stop_token = Event()
	def handle_signal(signum: int, frame):
		logger.log(LogFlag.SHOUT, f"caught {signal.Signals(signum).name}")
		stop_token.set()
	prev_handlers = {sig: signal.signal(sig, handle_signal) for sig in signals}
	try:
		yield stop_token
	finally:
		for sig, handler in prev_handlers.items():
			signal.signal(sig, handler)

class AbsWorker(Generic[T], AbsEventUnit, abstract=True):
	"""
	Fill results into pending :class:`T2 documents <ampel.content.T2Document.T2Document>`
	or :class:`T1 documents <ampel.content.T1Document.T1Document>`
	"""

	#: Ids of the units to run. If not specified, any t1/t2 unit will be run.
	unit_ids: None | list[str]

	#: Process only with the given code
	code_match: OneOrMany[int] = [DocumentCode.NEW, DocumentCode.RERUN_REQUESTED]

	#: Max number of docs to process in :func:`run`
	doc_limit: None | int

	#: Tag(s) to add to the stock :class:`~ampel.content.JournalRecord.JournalRecord`
	#: every time a document is processed
	jtag: None | OneOrMany[Tag]

	#: Tag(s) to add to the stock :class:`~ampel.content.MetaRecord.MetaRecord`
	#: every time a document is processed
	mtag: None | OneOrMany[Tag]

	#: Create a 'beacon' document indicating the last time T2Processor was run
	#: in the current configuration
	send_beacon: bool = True

	#: Explicitly call :func:`gc.collect` after the processing of each t1/t2 document.
	#: Turn this on if units do not clean up resources adequately,
	#: but beware of the performance impact.
	garbage_collect: bool = False

	#: Maximum number of processing attempts per document
	max_try: int = 5

	tier: ClassVar[Literal[1, 2]]

	#: Minimum age of documents to be matched (in seconds)
	#: {'$expr': {'$lt': [{'$last': '$meta.ts'}, now - min_doc_age]}}
	min_doc_age: None | float

	#: One day, we might support different DBs
	# database: str = "mongo"

	#: Avoid 'burning' a run_id for nothing at the cost of a request
	pre_check: bool = True

	#: Wait for documents to process
	backoff_on_query: None | BackoffConfig = None

	#: wait for majority acknowledgement of document updates. this introduces
	#: significant extra latency for replicated mongo clusters.
	wait_for_durable_write: bool = True

	#: read only commmitted updates
	durable_read: bool = False

	#: minimum number of stock document updates to commit at once
	updates_buffer_size: int = 500

	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)
		self._ampel_db = self.context.db
		self._loader = self.context.loader

		# Prepare DB query dict
		self.query: JDict = {
			'code': self.code_match if isinstance(self.code_match, DocumentCode)
				else maybe_match_array(self.code_match) # type: ignore[arg-type]
		}

		# Possibly restrict query to specified t2 units
		if self.unit_ids:
			self.query['unit'] = maybe_match_array(self.unit_ids)

		if self.min_doc_age:
			self.query['$expr'] = {
				'$lt': [{'$last': '$meta.ts'}, time() - self.min_doc_age]
			}

		# Shortcut
		self.col_stock = self._ampel_db.get_collection('stock', mode='r')
		self.col_t0 = self._ampel_db.get_collection('t0', mode='r')
		self.col_t1 = self._ampel_db.get_collection('t1', mode='r')
		self.col = self._ampel_db.get_collection(f't{self.tier}', mode='w')
		if not self.wait_for_durable_write:
			# Only wait for the primary to acknowledge (in-memory) write, rather
			# than a majority of replica set members to acknowledge write to
			# their journals. This has much lower latency, but can result in doc
			# updates being lost if the primary goes down before the write can
			# be replicated, but if this happens the doc can simply be rerun.
			self.col = self.col.with_options(write_concern=WriteConcern(w=1, j=False))
		if self.durable_read:
			self.col = self.col.with_options(read_concern=ReadConcern(level="majority"))

		if self.send_beacon:
			self.create_beacon()

		# _instances stores unit instances so that they can be re-used in run()
		# Key: set(unit name + config), value: unit instance
		self._instances: JDict = {}

		self._current_run_id: None | int = None


	@abstractmethod
	def process_doc(self,
		doc: T, ingester: AbsIngester, logger: AmpelLogger
	) -> Any:
		...


	def prepare(self, event_hdlr: EventHandler) -> None | EventCode:
		""" :returns: number of t2 docs processed """
		event_hdlr.set_tier(2)
		if self.pre_check and self.col.count_documents(self.query) == 0:
			return EventCode.PRE_CHECK_EXIT
		return None


	def find_one_and_update(self, query: dict, update: dict, stop_token: Event) -> None | T:
		"""
		Get next eligible document, with timeout
		"""
		t0 = time()
		attempts = 0
		while not stop_token.is_set():
			doc = self.col.find_one_and_update(
				query | {'$expr': {'$not': {'$lte': [ceil(time()), {'$last': '$meta.retry_after'}]}}},
				update
			)
			if doc is not None or self.backoff_on_query is None:
				return doc
			backoff = self.backoff_on_query
			if backoff.max_time is not None and time()-t0 > backoff.max_time:
				break
			if backoff.max_tries is not None and attempts+1 >= backoff.max_tries:
				break
			delay = min(backoff.factor*(backoff.base**attempts), backoff.max_value)
			if backoff.jitter:
				delay = random.uniform(0, delay)
			stop_token.wait(delay)
			attempts += 1
		
		return None


	def proceed(self, event_hdlr: EventHandler) -> int:
		""" :returns: number of t2 docs processed """

		event_hdlr.set_tier(self.tier)
		run_id = event_hdlr.get_run_id()

		if self.send_beacon:
			self.col_beacon.update_one(
				{'_id': self.beacon_id},
				{'$set': {'timestamp': int(time())}}
			)

		# Loop variables
		doc_counter = 0
		update = {'$set': {'code': DocumentCode.RUNNING}}
		garbage_collect = self.garbage_collect
		doc_limit = self.doc_limit

		with (
			AmpelLogger.from_profile(
				self.context, self.log_profile, run_id,
				base_flag = getattr(LogFlag, f'T{self.tier}') | LogFlag.CORE | self.base_log_flag
			) as logger,
			self._run_until_signal(run_id, logger) as stop_token,
 			MongoIngester(
				context = self.context, tier = self.tier, run_id = run_id,
				process_name = self.process_name, logger = logger,
				raise_exc = self.raise_exc, jtag = self.jtag
			) as ingester,
		):
			# Process docs until next() returns None (breaks condition below)
			while not stop_token.is_set():

				# get t1/t2 document (code is usually NEW or NEW_PRIO), excluding
				# docs with retry times in the future
				with stat_time.time() as timer:
					doc = self.find_one_and_update(self.query, update, stop_token)
					timer.labels(self.tier, "find_one_and_update", doc["unit"] if doc else None)

				# No match
				if doc is None:
					if not stop_token.is_set():
						logger.log(LogFlag.SHOUT, "No more docs to process")
					break
				if logger.verbose > 1:
					logger.debug(f'T{self.tier} doc to process', extra={'doc': doc})

				with (
					stat_time.labels(self.tier, "process_doc", doc["unit"]).time(),
					ingester.group()
				):
					self.process_doc(doc, ingester, logger)
				doc_counter += 1

				# Check possibly defined doc_limit
				if doc_limit and doc_counter >= doc_limit:
					break

				if garbage_collect:
					gc.collect()

		event_hdlr.add_extra(docs=doc_counter)

		return doc_counter


	@contextmanager
	def _run_until_signal(self,
		run_id: int,
		logger: AmpelLogger,
		signals: Sequence[signal.Signals] = (signal.SIGINT, signal.SIGTERM, signal.SIGQUIT, signal.SIGHUP)
	) -> Generator[Event, None, None]:
		self._current_run_id = run_id
		try:
			with stop_on_signal(signals, logger) as stop_token:
				yield stop_token
		finally:
			self._current_run_id = None
			self._instances.clear()
			self.context.loader.clear_adapter_cache()


	def _processing_error(self,
		logger: AmpelLogger, doc: T, body: UBson,
		meta: MetaRecord, msg: None | str = None,
		extra: None | JDict = None, exception: None | Exception = None
	) -> None:
		"""
		- Updates the t1/t2 document by appending a meta entry and
		updating 'code' with value DocumentCode.EXCEPTION if an exception
		was provided through the parameter 'exception' or with value
		DocumentCode.ERROR otherwise.

		- Updates the stock document by appending a JournalRecord to it
		and by updating the 'updated' timestamp

		- Creates a 'trouble' document in the troubles collection to report the incident

		:param msg: added to meta entry if provided
		"""

		if (
			not meta['activity'] or
			(len(meta['activity']) == 1 and meta['activity'][0]['action'] == 0)
		):
			meta['activity'] = []

		if not [el for el in meta['activity'] if (el['action'] & MetaActionCode.SET_CODE)]:
			meta['activity'].append( # type: ignore[attr-defined]
				{'action': MetaActionCode.SET_CODE}
			)

		if 'extra' not in meta:
			meta['extra'] = {}

		meta['extra']['msg'] = msg

		self.commit_update({'_id': doc['_id']}, meta, logger, code=DocumentCode.EXCEPTION) # type: ignore[typeddict-item]

		info: JDict = (extra or {}) | meta | {'stock': doc['stock'], 'doc': doc}
		if exception:
			report_exception(self._ampel_db, logger=logger, exc=exception, info=info)
		else:
			report_error(self._ampel_db, logger=logger, msg=msg, info=info)


	def commit_update(self,
		match: JDict,
		meta: MetaRecord,
		logger: AmpelLogger, *,
		payload_op: Literal['$push', '$set'] = '$push',
		body: UBson = None,
		tag: None | OneOrMany[Tag] = None,
		code: int = 0
	) -> None:
		"""
		Insert/upsert tier docs into DB.
		"""

		if logger.verbose:
			if logger.verbose > 1:
				if body is None:
					logger.log(DEBUG, "T2 unit returned no body")
				else:
					logger.log(DEBUG, f'Saving T{self.tier} unit result', extra={"body": body})
			else:
				logger.log(VERBOSE, f'Saving T{self.tier} unit result')

		upd: JDict = {
			'$set': {'code': code},
			'$push': {'meta': meta}
		}

		if self.mtag:

			tag = merge_tags(self.mtag, tag) if tag else self.mtag
			activities = meta['activity']

			# T2 unit added a tag, make the distinction clear by adding a dedicated activity
			if [el for el in activities if 'tag' in el]:
				activities.append( # type: ignore[attr-defined]
					{'action': MetaActionCode.ADD_WORKER_TAG, 'tag': self.mtag}
				)
			else:
				activities[0]['action'] |= MetaActionCode.ADD_WORKER_TAG
				activities[0]['tag'] = self.mtag

		if tag:
			upd['$addToSet'] = {
				'tag': tag if isinstance(tag, int | str) else maybe_use_each(tag)
			}

		if body is not None:
			upd[payload_op]['body'] = body

		# Update document
		self.col.update_one(match, upd)


	def gen_meta(self,
		run_id: int,
		unit_trace_id: None | int,
		duration: int | float,
		action_code: MetaActionCode = MetaActionCode(0) # noqa: B008
	) -> MetaRecord:

		d: MetaRecord = {
			'run': run_id,
			'ts': int(time()),
			'tier': self.tier,
			'code': None, # type: ignore[typeddict-item]
			'duration': duration,
			'activity': [{'action': action_code}],
			'traceid': {
				f't{self.tier}worker': self._trace_id,
				f't{self.tier}unit': unit_trace_id
			}
		}

		if self.job_sig:
			d['jobid'] = self.job_sig

		return d

	def create_beacon(self) -> None:
		""" Creates a beacon document if it does not exist yet """

		args = {
			'class': self.__class__.__name__,
			'unit': self.unit_ids,
			'code': self.code_match,
			'base_log_flag': self.base_log_flag.__int__(),
			'doc_limit': self.doc_limit
		}

		self.beacon_id = build_unsafe_dict_id(args)

		# Create beacon doc if it does not exist yet
		self.col_beacon = self._ampel_db.get_collection('beacon')

		# Create specific beacon doc if it does not exist yet
		self.col_beacon.update_one(
			{'_id': self.beacon_id},
			{
				'$setOnInsert': {
					'class': args.pop('class'),
					'config': convert_dollars(args)
				},
				'$set': {'timestamp': int(time())}
			},
			upsert=True
		)


	def get_unit_instance(self,
		doc: T1Document | T2Document,
		logger: AmpelLogger
	) -> LogicalUnit:

		k = f'{doc["unit"]}_{doc["config"]}'

		# Check if T2 instance exists in this run
		if k not in self._instances:

			# Instantiates unit with logger based on DefaultRecordBufferingHandler
			self._instances[k] = self.context.loader.new_safe_logical_unit(
				UnitModel(unit=doc['unit'], config=doc['config']),
				unit_type = LogicalUnit, # maybe restrict more in the future
				logger = logger,
				_chan = doc.get('channel') # type: ignore[arg-type] # probably not good to
			)

		return self._instances[k]


def register_stats(tier: int) -> tuple[Histogram, TimingCounter, Summary]:

	latency = AmpelMetricsRegistry.histogram(
		'latency',
		f'Delay between T{tier} doc creation and processing',
		subsystem=f't{tier}',
		unit='seconds',
		labelnames=('unit', ),
	)

	docs = AmpelMetricsRegistry.counter(
		'docs_processed',
		f'Number of T{tier} documents processed',
		subsystem=f't{tier}',
		labelnames=('unit', 'code',)
	)

	processing_time = AmpelMetricsRegistry.summary(
		'processing_time',
		f'Time spent processing of T{tier} documents processed',
		subsystem=f't{tier}',
		unit='microseconds',
		labelnames=('unit', 'phase')
	)

	return latency, docs, processing_time



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/test/dummy.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  11.02.2021
# Last Modified By:    jvs

from ampel.abstract.AbsStateT2Unit import AbsStateT2Unit


class DummyStateT2Unit(AbsStateT2Unit):

    foo: int = 42

    def process(self, compound, datapoints):
        return {"len": len(datapoints)}



import contextlib
import datetime
from collections import defaultdict
from collections.abc import Generator
from itertools import count
from typing import Any

import pytest
from pymongo.errors import DuplicateKeyError
from pytest_mock import MockerFixture, MockFixture

from ampel.abstract.AbsIngester import AbsIngester
from ampel.config.AmpelConfig import AmpelConfig
from ampel.content.DataPoint import DataPoint
from ampel.content.MetaRecord import MetaRecord
from ampel.content.T1Document import T1Document
from ampel.content.T2Document import T2Document
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.enum.DocumentCode import DocumentCode
from ampel.enum.MetaActionCode import MetaActionCode
from ampel.ingest.ChainedIngestionHandler import ChainedIngestionHandler, IngestBody
from ampel.log.AmpelLogger import DEBUG, AmpelLogger
from ampel.model.ingest.CompilerOptions import CompilerOptions
from ampel.model.ingest.IngestDirective import IngestDirective
from ampel.model.ingest.MuxModel import MuxModel
from ampel.model.ingest.T1Combine import T1Combine
from ampel.model.ingest.T2Compute import T2Compute
from ampel.model.UnitModel import UnitModel
from ampel.mongo.update.MongoIngester import MongoIngester
from ampel.queue.QueueIngester import AbsProducer, QueueIngester
from ampel.test.dummy import (
    DummyHistoryMuxer,
    DummyMuxer,
    DummyPointT2Unit,
    DummyStateT2Unit,
    DummyStockT2Unit,
)
from ampel.util.freeze import recursive_unfreeze


@pytest.fixture
def _dummy_units(dev_context: DevAmpelContext):
    for unit in (
        DummyStockT2Unit,
        DummyPointT2Unit,
        DummyStateT2Unit,
        DummyMuxer,
        DummyHistoryMuxer,
    ):
        dev_context.register_unit(unit)


@pytest.fixture(params=["T1SimpleCombiner", "T1SimpleRetroCombiner"])
def single_source_directive(
    dev_context: DevAmpelContext, _dummy_units, request
) -> IngestDirective:
    return IngestDirective(
        channel="TEST_CHANNEL",
        ingest=IngestBody(
            # https://github.com/python/mypy/issues/13421
            stock_t2=[T2Compute(unit="DummyStockT2Unit")],
            point_t2=[T2Compute(unit="DummyPointT2Unit")],
            combine=[
                T1Combine(
                    unit=request.param,
                    state_t2=[T2Compute(unit="DummyStateT2Unit")],
                )
            ],
        ),
    )


@pytest.fixture(params=["T1SimpleCombiner", "T1SimpleRetroCombiner"])
def multiplex_directive(
    dev_context: DevAmpelContext, _dummy_units, request
) -> IngestDirective:
    return IngestDirective(
        channel="TEST_CHANNEL",
        ingest=IngestBody(
            stock_t2=[T2Compute(unit="DummyStockT2Unit")],
            mux=MuxModel(
                unit="DummyMuxer",
                insert={"point_t2": [T2Compute(unit="DummyPointT2Unit")]},
                combine=[
                    T1Combine(
                        unit=request.param,
                        state_t2=[T2Compute(unit="DummyStateT2Unit")],
                    )
                ],
            ),
        ),
    )


def get_handler(
    context: DevAmpelContext,
    directives,
    ingester_model=UnitModel(unit="MongoIngester"),  # noqa: B008
) -> ChainedIngestionHandler:
    run_id = 0
    logger = AmpelLogger.get_logger(console={"level": DEBUG})
    ingester = context.loader.new_context_unit(
        ingester_model,
        context=context,
        sub_type=AbsIngester,
        logger=logger,
        run_id=run_id,
        tier=0,
        process_name="ingest",
        raise_exc=True,
        # need to disable provenance for dynamically registered unit
        _provenance=False,
    )
    return ChainedIngestionHandler(
        context=context,
        logger=logger,
        shaper=UnitModel(unit="NoShaper"),
        tier=0,
        trace_id={},
        run_id=run_id,
        compiler_opts=CompilerOptions(t0={"tag": ["TAGGERT"]}),
        ingester=ingester,
        directives=directives,
    )


def test_no_directive(dev_context):
    with pytest.raises(ValueError, match="Need at least 1 directive"):
        get_handler(dev_context, [])


@pytest.fixture
def datapoints() -> list[DataPoint]:
    return [
        {"id": i, "stock": "stockystock", "body": {"thing": i}}  # type: ignore[typeddict-item]
        for i in range(3)
    ]


def test_minimal_directive(
    dev_context: DevAmpelContext, datapoints: list[dict[str, Any]]
):
    """
    Minimal directive creates stock + t0 docs
    """
    directive = IngestDirective(channel="TEST_CHANNEL")
    handler = get_handler(dev_context, [directive])
    assert isinstance(handler.ingester, MongoIngester)

    # ingestion is idempotent
    handler.ingest(datapoints, [(0, True)], stock_id="stockystock")
    handler.ingester._updates_buffer.push_updates()
    handler.ingest(datapoints, [(0, True)], stock_id="stockystock")
    handler.ingester._updates_buffer.push_updates()

    stock = dev_context.db.get_collection("stock")
    assert stock.count_documents({}) == 1
    assert stock.count_documents({"stock": "stockystock"}) == 1

    for tier in range(4):
        col = dev_context.db.get_collection(f"t{tier}")
        assert col.count_documents({}) == (len(datapoints) if tier == 0 else 0)


def check_unit_counts(docs, num_states, num_points):
    assert len(docs) == 1 + num_states + num_points
    by_unit = defaultdict(list)
    for doc in docs:
        by_unit[doc["unit"]].append(doc)
    assert len(by_unit["DummyStateT2Unit"]) == num_states
    assert len(by_unit["DummyStockT2Unit"]) == 1
    assert len(by_unit["DummyPointT2Unit"]) == num_points
    for doc in docs:
        assert doc["stock"] == "stockystock"
        if doc["unit"] == "DummyStateT2Unit":
            assert isinstance(doc["link"], int)
        elif doc["unit"] == "DummyStockT2Unit":
            assert doc["link"] == "stockystock"
            assert doc["col"] == "stock"
        elif doc["unit"] == "DummyPointT2Unit":
            assert isinstance(doc["link"], int)
            assert doc["col"] == "t0"


def test_single_source_directive(
    dev_context, single_source_directive: IngestDirective, datapoints
):
    handler = get_handler(dev_context, [single_source_directive])
    assert isinstance(handler.ingester, MongoIngester)

    # include retro-completed states if configured
    assert single_source_directive.ingest.combine
    retro = "retro" in str(single_source_directive.ingest.combine[0].unit).lower()
    num_points = len(datapoints)
    num_states = num_points if retro else 1

    handler.ingest(datapoints, [(0, True)], stock_id="stockystock")
    handler.ingester._updates_buffer.push_updates()

    t0 = dev_context.db.get_collection("t0")
    t1 = dev_context.db.get_collection("t1")
    t2 = dev_context.db.get_collection("t2")

    assert t0.count_documents({}) == len(datapoints)
    assert t0.count_documents({"stock": "stockystock"}) == len(datapoints)
    assert t0.count_documents({"id": 0}) == 1

    assert t1.count_documents({}) == num_states if retro else 1

    check_unit_counts(list(t2.find({})), num_states, num_points)

    # test MongoT2Ingester idempotency
    before = t2.find_one_and_update({}, {"$set": {"code": DocumentCode.OK}})
    handler.ingest(datapoints, [(0, True)], stock_id="stockystock")
    handler.ingester._updates_buffer.push_updates()
    after = t2.find_one({"_id": before["_id"]})
    assert before["code"] == DocumentCode.NEW
    assert after["code"] == DocumentCode.OK, "code was not overwitten"
    assert before["body"] == after["body"]
    assert len(after["meta"]) > len(before["meta"])


def test_queue_ingester(
    dev_context,
    single_source_directive: IngestDirective,
    datapoints,
    mocker: MockerFixture,
):
    @dev_context.register_unit
    class DummyProducer(AbsProducer):
        def __init__(self, **kwargs) -> None:
            super().__init__(**kwargs)
            self.items: list[AbsProducer.Item] = []

        def produce(self, item: AbsProducer.Item, delivery_callback=None):
            self.items.append(item)
            if delivery_callback:
                delivery_callback()

        def __exit__(self, exc_type, exc_value, traceback) -> None:
            return

    acknowledge_callback = mocker.MagicMock()

    dev_context.register_unit(QueueIngester)
    handler = get_handler(
        dev_context,
        [single_source_directive],
        UnitModel(
            unit="QueueIngester",
            config={
                "producer": {
                    "unit": "DummyProducer",
                },
                "acknowledge_callback": acknowledge_callback,
            },
        ),
    )
    assert isinstance(handler.ingester, QueueIngester)
    assert isinstance(handler.ingester._producer, DummyProducer)

    # include retro-completed states if configured
    assert single_source_directive.ingest.combine
    retro = "retro" in str(single_source_directive.ingest.combine[0].unit).lower()
    num_points = len(datapoints)
    num_states = num_points if retro else 1

    sentinel = object()

    with handler.ingester.group([sentinel]):
        handler.ingest(datapoints, [(0, True)], stock_id="stockystock")
    assert len(items := handler.ingester._producer.items) == 1
    assert acknowledge_callback.call_count == 1, "acknowledge callback called"
    assert list(acknowledge_callback.call_args[0][0]) == [sentinel], "callback payload contains sentinel"
    with handler.ingester.group():
        handler.ingest(datapoints, [(0, True)], stock_id="stockystock")
    assert len(items := handler.ingester._producer.items) == 2, (
        "second ingestion creates a second message"
    )
    assert acknowledge_callback.call_count == 1, "acknowledge callback called only once"

    t0 = items[0].t0
    t1 = items[0].t1
    t2 = items[0].t2

    assert len(t0) == len(datapoints)
    assert all(doc["stock"] == "stockystock" for doc in t0)
    assert len([d for d in t0 if d["id"] == 0]) == 1

    assert len(t1) == num_states if retro else 1

    check_unit_counts(t2, num_states, num_points)


def test_multiplex_directive(
    dev_context, multiplex_directive: IngestDirective, datapoints
):
    handler = get_handler(dev_context, [multiplex_directive])
    assert isinstance(handler.ingester, MongoIngester)

    # include retro-completed states if configured
    assert multiplex_directive.ingest.mux
    assert multiplex_directive.ingest.mux.combine
    retro = "retro" in str(multiplex_directive.ingest.mux.combine[0].unit).lower()
    num_points = len(datapoints) + 5
    num_states = num_points if retro else 1

    handler.ingest(datapoints, [(0, True)], stock_id="stockystock")
    handler.ingester._updates_buffer.push_updates()

    t0 = dev_context.db.get_collection("t0")
    t1 = dev_context.db.get_collection("t1")
    t2 = dev_context.db.get_collection("t2")

    assert t0.count_documents({}) == num_points
    assert t0.count_documents({"stock": "stockystock"}) == num_points
    assert t0.count_documents({"id": 0}) == 1

    assert t1.count_documents({}) == num_states if retro else 1

    check_unit_counts(list(t2.find({})), num_states, num_points)


def test_multiplex_dispatch(
    dev_context,
    single_source_directive: IngestDirective,
    multiplex_directive: IngestDirective,
    datapoints,
):
    """
    Extended history compounds are created only for channels that request it
    """
    single_source_directive.channel = "TEST_CHANNEL"
    multiplex_directive.channel = "LONG_CHANNEL"

    handler = get_handler(dev_context, [single_source_directive, multiplex_directive])
    assert isinstance(handler.ingester, MongoIngester)

    # include retro-completed states if configured
    assert single_source_directive.ingest.combine
    assert single_source_directive.ingest.combine[0]
    assert multiplex_directive.ingest.mux
    assert multiplex_directive.ingest.mux.combine
    assert multiplex_directive.ingest.mux.combine[0]
    retro_short = "retro" in str(single_source_directive.ingest.combine[0].unit).lower()
    retro_long = "retro" in str(multiplex_directive.ingest.mux.combine[0].unit).lower()
    num_points = len(datapoints) + 5

    handler.ingest(datapoints, [(0, True), (1, True)], stock_id="stockystock")
    handler.ingester._updates_buffer.push_updates()

    t0 = dev_context.db.get_collection("t0")
    t1 = dev_context.db.get_collection("t1")

    assert t0.count_documents({"channel": "TEST_CHANNEL"}) == len(datapoints), (
        "single-stream channel contains only direct datapoints"
    )
    assert t0.count_documents({"channel": "LONG_CHANNEL"}) == num_points, (
        "multiplexed channel contains additional datapoints"
    )

    assert t1.count_documents({"channel": "TEST_CHANNEL"}) == (
        len(datapoints) if retro_short else 1
    ), "single-stream channel has the correct number of states"
    assert t1.count_documents({"channel": "LONG_CHANNEL"}) == (
        num_points if retro_long else 1
    ), "single-stream channel has the correct number of states"
    # if retro-completion settings are different, channels' states are disjoint
    assert (
        t1.count_documents(
            {"$and": [{"channel": "TEST_CHANNEL"}, {"channel": "LONG_CHANNEL"}]}
        )
        == 0
    ), "channel states are disjoint"


def test_multiplex_elision(
    dev_context,
    single_source_directive: IngestDirective,
    multiplex_directive: IngestDirective,
    datapoints,
):
    """
    Extended history points are skipped when only short channels pass
    """
    single_source_directive.channel = "TEST_CHANNEL"
    multiplex_directive.channel = "LONG_CHANNEL"

    handler = get_handler(dev_context, [single_source_directive, multiplex_directive])
    assert isinstance(handler.ingester, MongoIngester)

    handler.ingest(datapoints, [(0, True)], stock_id="stockystock")
    handler.ingester._updates_buffer.push_updates()

    t0 = dev_context.db.get_collection("t0")

    assert t0.count_documents({"channel": "TEST_CHANNEL"}) == len(datapoints), (
        "single-stream channel contains only direct datapoints"
    )
    assert t0.count_documents({"channel": "LONG_CHANNEL"}) == 0, (
        "multiplexed channel contains additional datapoints"
    )


def test_t0_meta_append(
    mock_context: DevAmpelContext,
    datapoints: list[DataPoint],
):
    """T0Compiler preserves tags and meta entries"""
    handler = get_handler(mock_context, [IngestDirective(channel="TEST_CHANNEL")])
    assert isinstance(handler.ingester, MongoIngester)
    ts = 3.14159
    meta_record: MetaRecord = {
        "activity": [
            {
                "code": MetaActionCode.ADD_INGEST_TAG,
            }
        ],
        "ts": ts,
        "extra": {"thing": "floop"},
    }
    tags = ["SOME_TAG", "SOME_OTHER_TAG"]

    datapoints[0]["meta"] = [meta_record]
    datapoints[0]["tag"] = tags
    handler.t0_compiler.add(datapoints, "SOME_CHANNEL", None, trace_id=0)
    handler.t0_compiler.commit(handler.ingester.t0, ts)
    handler.ingester._updates_buffer.push_updates(force=True)

    doc = mock_context.db.get_collection("t0").find_one({"id": datapoints[0]["id"]})
    assert doc is not None
    assert doc["channel"] == ["SOME_CHANNEL"]
    assert set(tags).intersection(doc["tag"]), "initial datapoint tags set"
    assert set(doc["tag"]).difference(tags), "compiler adds its own tags"
    assert len(doc["meta"]) > 1, "initial datapoint meta set"
    assert meta_record in doc["meta"], "compiler adds its own meta entries"


@contextlib.contextmanager
def unfreeze_config(context: DevAmpelContext) -> Generator[dict, None, None]:
    config = context.config
    writable_config = recursive_unfreeze(config.get())  # type: ignore[arg-type]
    context.config = AmpelConfig(writable_config, freeze=False)
    yield writable_config
    context.config = config


def test_duplicate_t0_id(
    integration_context: DevAmpelContext,
    datapoints: list[DataPoint],
):
    def run() -> None:
        handler = get_handler(
            integration_context, [IngestDirective(channel="TEST_CHANNEL")]
        )
        assert isinstance(handler.ingester, MongoIngester)
        handler.t0_compiler.add(datapoints, "SOME_CHANNEL", ttl=None, trace_id=0)
        handler.t0_compiler.commit(handler.ingester.t0, 0)
        handler.ingester._updates_buffer.push_updates(force=True)

    # populate documents
    run()

    # fake an id collision by changing the body of an existing document without
    # updating its id. when the consumer runs again below, this should cause the
    # update operation to fail to match the body, and attempt to insert a duplicate
    # id
    datapoints[0]["body"]["broken"] = "boom"
    with unfreeze_config(integration_context) as config:
        config["mongo"]["ingest"]["t0"] = {
            "unit": "MongoT0Ingester",
            "config": {"extended_match": 2},
        }
        with pytest.raises(DuplicateKeyError):
            run()

    # with id check disabled, no exception is raised
    run()


def test_t0_ttl(
    mock_context: DevAmpelContext,
    datapoints: list[DataPoint],
    mocker: MockFixture,
):
    """
    Previously inserted datapoint ttls are updated when used by a muxer
    """
    mocker.patch(
        "ampel.ingest.ChainedIngestionHandler.time", side_effect=count().__next__
    )
    mock_context.register_units(DummyHistoryMuxer)

    ttl = datetime.timedelta(seconds=300)
    with unfreeze_config(mock_context) as config:
        config["mongo"]["ingest"]["t0"] = {
            "unit": "MongoT0Ingester",
            "config": {"update_ttl": True},
        }
        config["channel"]["TEST_CHANNEL"]["purge"]["content"]["delay"] = {
            "seconds": ttl.total_seconds()
        }

        handler = get_handler(
            mock_context,
            [
                IngestDirective(
                    channel="TEST_CHANNEL",
                    ingest={  # type: ignore[arg-type]
                        "mux": {
                            "unit": "DummyHistoryMuxer",
                            "combine": [
                                {
                                    "unit": "T1SimpleCombiner",
                                    "state_t2": [{"unit": "DummyStateT2Unit"}],
                                }
                            ],
                        }
                    },
                )
            ],
        )

    def ingest(datapoints: list[DataPoint]):
        handler.ingest(
            [dict(dp) for dp in datapoints], [(0, True)], stock_id="stockystock"
        )
        assert isinstance(handler.ingester, MongoIngester)
        handler.ingester._updates_buffer.push_updates(force=True)

    def get_meta_time(dp: DataPoint | T1Document | T2Document) -> datetime.datetime:
        return datetime.datetime.fromtimestamp(
            dp["meta"][-1]["ts"], tz=datetime.timezone.utc
        )

    def get_expire_time(dp: DataPoint | T1Document | T2Document):
        # NB: mongo datetimes are implicitly utc, so reattach tzinfo
        return dp["expiry"].replace(tzinfo=datetime.timezone.utc)

    ingest(datapoints[:1])
    for tier in range(3):
        assert (
            collection := mock_context.db.get_collection(f"t{tier}")
        ).count_documents({}) == 1, f"t{tier} document inserted"
        doc: None | DataPoint | T1Document | T2Document = collection.find_one()
        assert doc is not None
        assert get_expire_time(doc) == get_meta_time(doc) + ttl, (
            f"ttl set on t{tier} document"
        )

    # ingest remaining datapoints
    ingest(datapoints)
    dp: None | DataPoint = mock_context.db.get_collection("t0").find_one({"id": 0})
    assert dp is not None
    assert get_expire_time(dp) > get_meta_time(dp) + ttl, (
        f"ttl updated for dp {dp['id']}"
    )
    assert len(dp["meta"]) == 1, "no extra meta entries added"

    # and again, ensuring that all ttls get updated
    ingest(datapoints)
    for dp in mock_context.db.get_collection("t0").find():
        assert dp is not None
        assert get_expire_time(dp) > get_meta_time(dp) + ttl, (
            f"ttl updated for dp {dp['id']}"
        )
        assert len(dp["meta"]) == 1, "no extra meta entries added"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsDocUpdater.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                28.05.2021
# Last Modified Date:  10.10.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any, Literal

from ampel.base.AmpelABC import AmpelABC
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.base.decorator import abstractmethod


class AbsDocUpdater(AmpelABC, AmpelBaseModel, abstract=True):
	"""
	Aim at facilating updates of documents regardless of the underlying database system.
	Not in use yet
	"""

	@abstractmethod
	def update(self,
		tier: Literal[0, 1, 2],
		match: dict[str, Any],
		let: dict[str, Any],
		push: dict[str, Any],
		add_to_set: dict[str, Any]
	) -> None:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/collector/AbsConfigCollector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.10.2019
# Last Modified Date:  03.03.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from string import Template
from typing import Any, Literal

from ampel.config.builder.DisplayOptions import DisplayOptions
from ampel.log.AmpelLogger import AmpelLogger


class ConfigCollector(dict):

	def __init__(self,
		conf_section: str,
		options: DisplayOptions | None = None,
		content: dict | None = None,
		logger: AmpelLogger | None = None,
		tier: Literal[0, 1, 2, 3, "ops"] | None = None,
		ignore_exc: list[str] | None = None
	) -> None:

		super().__init__(**content if content else {})
		self.options = options or DisplayOptions()
		self.verbose = self.options.verbose
		self.has_error = False
		self.conf_section = conf_section
		self.logger = AmpelLogger.get_logger() if logger is None else logger
		self.tier = f'T{tier}' if tier is not None else 'general'
		self.ignore_exc = ignore_exc

		if self.options.debug:
			self.logger.info(
				f'Creating {self.__class__.__name__} '
				f'for {self.tier} config section "{conf_section}"'
			)


	def error(self, msg: str, exc_info: None | Any = None) -> None:
		self.logger.error(msg, exc_info=exc_info)
		self.has_error = True


	def missing_key(self,
		what: str, key: str, dist_name: str, register_file: None | str
	) -> None:
		self.error(
			f'{what} dict is missing key "{key}" '
			f'{self.distrib_hint(dist_name, register_file)}'
		)


	def report_duplicated_entry(self,
		conf_key: str,
		prev_file: str,
		prev_dist: str,
		new_file: str,
		new_dist: str,
		section_detail: None | str = None
	) -> None:

		t = Template(
			'Duplicated $what definition: "$conf_key"\n' +
			' Previously defined by $prev\n' +
			' Redefined by $new'
		)

		prev = self.distrib_hint(prev_dist, prev_file)
		new = self.distrib_hint(new_dist, new_file)

		self.error(
			t.substitute(
				what = section_detail if section_detail else self.conf_section,
				conf_key = conf_key,
				prev = prev,
				new = new if new else 'unknown',
			)
		)


	@staticmethod
	def distrib_hint(
		distrib: None | str = None,
		file_register: None | str = None
	) -> str:
		""" Adds distribution name if available """

		ret = ''

		if file_register:
			ret += file_register

		if distrib:
			if ret:
				return f'{ret} ({distrib})'
			return f'({distrib})'

		if ret:
			return f'{ret})'

		return ret



import copy

from ampel.config.alter.HashT2Config import HashT2Config
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.test.dummy import DummyStateT2Unit, DummyTiedStateT2Unit


def test_recursive_hash(mock_context: DevAmpelContext, ampel_logger):
    """
    HashT2Config.alter should hash nested t2 configs
    """
    unit_config = {
        "t2_dependency": [{"unit": "DummyStateT2Unit", "config": {"foo": 37}}]
    }
    for u in (DummyStateT2Unit, DummyTiedStateT2Unit):
        mock_context.register_unit(u)

    hashed = HashT2Config().alter(
        mock_context,
        dict(
            directives=[
                dict(
                    channel="TEST",
                    ingest=dict(
                        combine=[
                            dict(
                                unit="T1SimpleCombiner",
                                state_t2=[
                                    dict(
                                        unit="DummyTiedStateT2Unit",
                                        config=copy.deepcopy(unit_config),
                                    )
                                ],
                            )
                        ]
                    ),
                )
            ]
        ),
        ampel_logger,
    )

    toplevel_config_id = hashed["directives"][0]["ingest"]["combine"][0]["state_t2"][0][
        "config"
    ]
    assert isinstance(toplevel_config_id, int)
    toplevel_config = mock_context.get_config().get_conf_by_id(toplevel_config_id)
    nested_config_id = toplevel_config["t2_dependency"][0]["config"]
    assert isinstance(nested_config_id, int)
    assert (
        mock_context.get_config().get_conf_by_id(nested_config_id)
        == unit_config["t2_dependency"][0]["config"]
    )



from ampel.abstract.AbsT0Muxer import AbsT0Muxer
from ampel.content.DataPoint import DataPoint
from ampel.model.UnitModel import UnitModel
from ampel.types import DataPointId, StockId


class ChainedT0Muxer(AbsT0Muxer):
    """
    A sequence of muxers
    """

    muxers: list[UnitModel]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._muxers = [
            self.context.loader.new_context_unit(
                model=model, logger=self.logger, sub_type=AbsT0Muxer, context=self.context,
                updates_buffer=self.updates_buffer
            )
            for model in self.muxers
        ]

    def process(
        self, dps: list[DataPoint], stock_id: None | StockId = None
    ) -> tuple[None | list[DataPoint], None | list[DataPoint]]:
        """
        :returns: (to_insert,to_combine), where to_insert is the union of the insertable points
          returned by each muxer, and to_combine is the final datapoint sequence
        """
        
        dps_to_insert: dict[tuple[DataPointId, None | int], DataPoint] = dict()
        dps_to_combine: None | list[DataPoint] = dps
        
        for muxer in self._muxers:
            insert, dps_to_combine = muxer.process(dps_to_combine, stock_id)
            for dp in (insert or []):
                dps_to_insert[(dp["id"], dp.get("origin"))] = dp
        
        return list(dps_to_insert.values()) or None, dps_to_combine



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/core/Schedulable.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                26.05.2018
# Last Modified Date:  19.04.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import logging
import signal
import threading
from collections.abc import Callable
from contextlib import contextmanager
from logging import Logger

from schedule import Scheduler


class Schedulable:
	"""
	Class relying on module 'schedule' for scheduling jobs at given times.
	The permanent loop required by the scheduler runs within method "run_scheduler".
	The latter can be started in its own thread using start() and stopped using stop().
	The scheduler loop can also run in a context using run_in_thread() which should allow
	the python interpreter to exit after an unhandled exception.
	"""

	def __init__(self,
		start_callback: None | Callable = None,
		stop_callback: None | Callable = None,
		catch_signals: None | bool = True
	) -> None:
		"""
		:param start_callback: method to be executed before starting the run_scheduler thread.
		:param stop_callback: method to be executed after run_scheduler thread join().
		"""

		logging.getLogger('schedule').propagate = False
		logging.getLogger('schedule').setLevel(logging.WARNING)

		if catch_signals:
			# Catch term/int signals
			signal.signal(signal.SIGINT, self.sig_exit)
			signal.signal(signal.SIGTERM, self.sig_exit)

		# Callback functions for start and stop methods
		self._start_callback = start_callback
		self._stop_callback = stop_callback

		# Create scheduler
		self._scheduler = Scheduler()
		self._event = threading.Event()


	def get_scheduler(self) -> Scheduler:
		""" Returns Scheduler instance associated with this class instance """
		return self._scheduler


	@contextmanager
	def run_in_thread(self, logger: None | Logger = None):
		"""
		Runs 'run_scheduler' in its own thread using contextmanager.
		The methods start() and stop() are used under the hood.
		"""

		self.start(logger)
		try:
			yield self
		finally:
			self.stop()


	def start(self, logger: None | Logger = None) -> None:
		"""
		Executes method 'run_scheduler()' in its own thread.
		If start_callback was provided as argument in constructor,
		_start_callback() will be executed before run_scheduler()
		"""
		self._event.clear() # sets event flag to false

		if self._start_callback:
			self._start_callback()

		self._thread_run = threading.Thread(
			target=self.run_scheduler, args=(logger, )
		)

		self._thread_run.start()


	def stop(self) -> None:
		"""
		Stops run_scheduler() thread.
		If stop_callback was provided as argument in constructor,
		_stop_callback() will be executed after run_scheduler() thread joins
		"""

		self._event.set() # sets event flag to true

		if hasattr(self, '_thread_run'):
			self._thread_run.join()

		if self._stop_callback:
			self._stop_callback()


	def sig_exit(self, signum: int, frame) -> None:
		""" Executed when SIGTERM/SIGINT is caught. Stops the run_scheduler() thread. """
		self._event.set() # Sets event flag to true


	def run_scheduler(self, logger: None | Logger = None) -> None:
		"""
		Runs scheduler main loop.
		See https://schedule.readthedocs.io
		"""

		if logger:
			logger.info(f"Starting {self.__class__.__name__} scheduler loop")

		# wait() method blocks until event flag _event is true,
		# that is until _event.set() is called
		while not self._event.wait(1):
			self._scheduler.run_pending()

		if logger:
			logger.info(f"{self.__class__.__name__} scheduler loop stopped")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/t3/T2FilterModel.py
# License:             BSD-3-Clause
# Author:              Jakob van Santen <jakob.van.santen@desy.de>
# Date:                02.08.2020
# Last Modified Date:  02.08.2020
# Last Modified By:    Jakob van Santen <jakob.van.santen@desy.de>

from typing import Any

from ampel.base.AmpelBaseModel import AmpelBaseModel


class T2FilterModel(AmpelBaseModel):

	#: name of T2 unit
	unit: str
	#: Mongo match expression to apply to most recent result
	match: dict[str, Any]


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsT0Muxer.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                10.05.2021
# Last Modified Date:  13.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.content.DataPoint import DataPoint
from ampel.core.ContextUnit import ContextUnit
from ampel.log.AmpelLogger import AmpelLogger
from ampel.types import StockId, Traceless


class AbsT0Muxer(AmpelABC, ContextUnit, abstract=True):
	"""
	Combines a main source of datapoints with one or more additional sources.
	Unlike t1 combiner, subclass of this unit can fetch and add additional points (possibly from different instruments)
	"""

	logger: Traceless[AmpelLogger]

	@abstractmethod
	def process(self,
		dps: list[DataPoint],
		stock_id: None | StockId = None
	) -> tuple[None | list[DataPoint], None | list[DataPoint]]:
		"""
		Potentially:
		- Append datapoints to the datapoints provided as argument (the source can be the AmpelDB or external source(s))
		- Reduce to the number of datapoints to insert by determining which datapoints are already in the DB

		Subclasses of this class are used by the ingestion handler,
		which usually performs the following sequence of operation:
		1) shape (T0 unit, must be done first to make sure datapoint ids are avail for comparison)
		2) t0 complete (subclass of this class. For example DB-completition)
		3) ingest
		4) t1 combine

		:returns: tuple[datapoints to insert, datapoints to combine]
		<datapoints to insert> will be provided to a T0 ingester
		<datapoints to combine> will potentially be provided to an underlying T1 combiner

		Notes regarding <datapoints to combine>:
		- may contain the union between DB datapoints and provided datapoints (dps)
		- it may not contain all datapoints (potiential superseded db datapoints may be excluded)
		- If projections are used by subclass (for optimization purposes),
		  not all the datapoints may contain the same information (some might be missing non-projected keys)
		"""
		raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/builder/ConfigChecker.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                03.09.2019
# Last Modified Date:  16.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>


import json
import os
import sys
import traceback
from typing import Any

from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.config.builder.BaseConfigChecker import BaseConfigChecker
from ampel.core.AmpelContext import AmpelContext
from ampel.model.UnitModel import UnitModel
from ampel.mongo.update.DBUpdatesBuffer import DBUpdatesBuffer
from ampel.secret.AmpelVault import AmpelVault
from ampel.secret.PotemkinSecretProvider import PotemkinSecretProvider
from ampel.util.pretty import prettyjson
from ampel.util.recursion import walk_and_process_dict


class ConfigChecker(BaseConfigChecker):
	"""
	Validate a config by using it to initialize an AmpelContext, and then
	instantiating all processors and the units defined in their configurations.
	"""

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.ctx = AmpelContext.load(
			self.config,
			vault=AmpelVault(providers=[PotemkinSecretProvider()]),
		)


	def validate(self,
		ignore_inactive: bool = False,
		ignore_ressource_not_avail: bool = True,
		raise_exc: bool = False
	) -> dict[str, Any]:
		"""
		:returns: config if check passed
		:raises: BadConfig
		"""

		self.validate_processor_models(
			ignore_inactive, ignore_ressource_not_avail, raise_exc
		)

		# Recursively validate all UnitModels
		self.validate_unit_models(
			ignore_inactive, ignore_ressource_not_avail, raise_exc
		)

		return self.config


	def load_model(self,
		tier: str, proc: str, load_callable: Any,
		model_args: dict[str, Any], raise_exc: bool = False,
		ignore_ressource_not_avail: bool = False,
		load_args: None | dict[str, Any] = None
	) -> bool:
		"""
		:param proc: super process name
		:returns: True on error, False otherwise
		"""

		try:
			# block potential "Offending value..." print from AmpelBaseModel
			# since we pretty print the input values in case of errors
			sys.stdout = open(os.devnull, 'w') # noqa: SIM115
			load_callable(**(model_args | load_args if load_args else model_args))
	
		except Exception as e:

			if ignore_ressource_not_avail and "Global resource not available" in str(e):
				self.logger.error("Ignoring ressource not available error", extra={"process": proc})
			else:

				if raise_exc:
					raise e

				self.logger.error("Processor model validation error", extra={"process": proc})
				for el in prettyjson(model_args).split("\n"):
					self.logger.error(el, extra={"process": proc})

				self._log_exc(e, proc)
				return True

		finally:
			# Restore print
			sys.stdout.close()
			sys.stdout = sys.__stdout__

		return False


	def validate_processor_models(self,
		ignore_inactive: bool = False,
		ignore_ressource_not_avail: bool = False,
		raise_exc: bool = False
	) -> None:

		for tier, proc in self.iter_procs(ignore_inactive):

			if self.verbose:
				self.logger.info("Validating processor model", extra={"process": proc})

			if self.load_model(
				tier = tier, proc = proc,
				load_callable = self.loader.new_context_unit,
				load_args = {"context": self.ctx},
				model_args = {
					"model": UnitModel(
						unit=self.config['process'][tier][proc]['processor']['unit'],
						config={
							**self.config['process'][tier][proc]['processor'].get('config', {}),
							"process_name": proc
						}
					),
				},
				raise_exc = raise_exc,
				ignore_ressource_not_avail = ignore_ressource_not_avail,
			):
				del self.config['process'][tier][proc]


	def validate_unit_models(self,
		ignore_inactive: bool = False,
		ignore_ressource_not_avail: bool = False,
		raise_exc: bool = False
	) -> None:

		for tier, proc in self.iter_procs(ignore_inactive):

			unit_models: list[dict[str, Any]] = []
			walk_and_process_dict(
				arg = self.config['process'][tier][proc]['processor'].get('config'),
				callback = self._gather_unit_models_callback,
				match = ['config'],
				unit_models = unit_models
			)

			for um in unit_models:

				try:

					if self.verbose:
						self.logger.debug(f"Checking model {um['path']}")

					if um['model']['unit'] in AuxUnitRegister._defs:  # noqa: SLF001
						self.load_model(
							tier = tier, proc = proc,
							load_callable = AuxUnitRegister.new_unit,
							model_args = self._customize_aux_models(um),
							raise_exc = raise_exc,
							ignore_ressource_not_avail = ignore_ressource_not_avail,
						)
					elif um['model']['unit'] in self.config["unit"]["admin"]:
						# Deactivated for now, too complicated
						self.load_model(
							tier = tier, proc = proc,
							load_callable = self.loader.new_context_unit,
							load_args = {"context": self.ctx},
							model_args = self._customize_admin_models(um),
							raise_exc = raise_exc,
							ignore_ressource_not_avail = ignore_ressource_not_avail,
						)
					elif um['model']['unit'] in self.config["unit"]["base"]:
						self.load_model(
							tier = tier, proc = proc,
							load_callable = self.loader.new_logical_unit,
							load_args = {"logger": self.logger},
							model_args = um["model"],
							raise_exc = raise_exc,
							ignore_ressource_not_avail = ignore_ressource_not_avail,
						)

				except Exception as e:  # noqa: PERF203

					if raise_exc:
						raise e

					self.logger.error("Unit model validation error", extra={"process": proc})
					for el in prettyjson(um['model']).split("\n"):
						self.logger.error(el, extra={"process": proc})
					self._log_exc(e, proc)


	def _customize_admin_models(self, um: dict[str, Any]) -> dict[str, Any]:

		model = json.loads(json.dumps(um['model']))
		unit_name = model['unit']
		if "config" not in model or model["config"] is None:
			model["config"] = {}

		for k in list(model.keys()):
			# Ingester extra keys
			if k in ["units", "t2_compute", "t1_combine", "t3_supervize"]:
				model.pop(k)

		if "AbsDocIngester" in self.config["unit"]["admin"][unit_name]["base"]:
			model["config"].update({
				"run_id": 1, "logd": {"logs": [], "extra": {}, "err": False},
				"updates_buffer": DBUpdatesBuffer(self.ctx.db, run_id=1, logger=self.logger)
			})

		elif (
			"AbsT3Stager" in self.config["unit"]["admin"][unit_name]["base"] or
			"AbsT3Supplier" in self.config["unit"]["admin"][unit_name]["base"]
		):
			model["config"].update({"process_name": "test", "run_id": 1, "logger": self.logger})

		elif "AbsT3Selector" in self.config["unit"]["admin"][unit_name]["base"]:
			model["config"]["logger"] = self.logger

		return model


	def _customize_aux_models(self, um: dict[str, Any]) -> dict[str, Any]:

		model = um['model']
		unit = model['unit']

		if (
			unit in self.config["unit"]["aux"] and
			"AbsT3Projector" in self.config["unit"]["aux"][unit]["base"]
		):
			model = json.loads(json.dumps(model))
			if "config" not in model or model["config"] is None:
				model["config"] = {}
			model["config"]["logger"] = self.logger

		return model


	def _gather_unit_models_callback(self, path, k, d, **kwargs) -> None:

		if self.verbose:
			self.logger.debug(f"# path: {path}.config")

		kwargs['unit_models'].append(
			{'path': path, 'model': d}
		)


	def _log_exc(self, e: Exception, proc: str) -> None:
		self.logger.error("=" * 80, extra={"process": proc})
		for el in traceback.format_exception(type(e), e, e.__traceback__):
			for ell in el.split('\n'):
				if len(ell) > 0:
					self.logger.error(ell, extra={"process": proc})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/core/UnitLoader.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                07.10.2019
# Last Modified Date:  11.11.2025
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import contextlib
import os
import sys
from collections.abc import Iterator
from contextlib import contextmanager
from hashlib import blake2b
from importlib import import_module
from pathlib import Path
from typing import Any, TypeVar, overload

from ampel.base.AmpelUnit import AmpelUnit
from ampel.base.AuxUnitRegister import AuxUnitRegister
from ampel.base.LogicalUnit import LogicalUnit
from ampel.config.AmpelConfig import AmpelConfig
from ampel.core.AmpelContext import AmpelContext
from ampel.core.AmpelDB import AmpelDB
from ampel.core.ContextUnit import ContextUnit
from ampel.abstract.AbsUnitResultAdapter import AbsUnitResultAdapter
from ampel.log.AmpelLogger import VERBOSE, AmpelLogger, LogFlag
from ampel.log.handlers.ChanRecordBufHandler import ChanRecordBufHandler
from ampel.log.handlers.DefaultRecordBufferingHandler import DefaultRecordBufferingHandler
from ampel.model.t3.AliasableModel import AliasableModel
from ampel.model.UnitModel import UnitModel
from ampel.secret.AmpelVault import AmpelVault
from ampel.secret.NamedSecret import NamedSecret
from ampel.types import ChannelId, check_class
from ampel.util.collections import ampel_iter
from ampel.util.freeze import recursive_unfreeze
from ampel.util.hash import build_unsafe_dict_id
from ampel.util.mappings import merge_dicts
from ampel.view.ReadOnlyDict import ReadOnlyDict

T = TypeVar('T', bound=AmpelUnit)
LT = TypeVar('LT', bound=LogicalUnit)
CT = TypeVar('CT', bound=ContextUnit)
pyv = f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
env = ('conda_' + os.environ["CONDA_DEFAULT_ENV"]) if 'CONDA_DEFAULT_ENV' in os.environ else 'default'

class UnitLoader:

	def __init__(self,
		config: AmpelConfig,
		db: None | AmpelDB,
		provenance: bool = True,
		vault: None | AmpelVault = None
	) -> None:
		"""
		:raises: ValueError in case bad arguments are provided
		"""

		if not isinstance(config, AmpelConfig):
			raise ValueError(
				f"First parameter must be an instance of "
				f"AmpelConfig (provided: {type(config)})"
			)
		
		if provenance and not db:
			raise ValueError("Provenance tracking requires a database connection")

		self.db = db
		self.vault = vault
		self.config = config
		self.provenance = provenance
		self.unit_defs: dict = config._config['unit']  # noqa: SLF001
		self.aliases: list[dict] = [config._config['alias'][f"t{el}"] for el in (0, 3, 1, 2)]  # noqa: SLF001

		# potentially updated by DevAmpelContext
		self._dyn_register: dict[str, type[LogicalUnit] | type[ContextUnit]] | None = None
		self._adapters: dict[str, AbsUnitResultAdapter] = {}


	@overload
	def new_logical_unit(self,
		model: UnitModel, logger: AmpelLogger, *, sub_type: type[LT], **kwargs
	) -> LT:
		...
	@overload
	def new_logical_unit(self,
		model: UnitModel, logger: AmpelLogger, *, sub_type: None = ..., **kwargs
	) -> LogicalUnit:
		...
	def new_logical_unit(self,
		model: UnitModel,
		logger: AmpelLogger, *,
		sub_type: None | type[LT] = None,
		**kwargs
	) -> LT | LogicalUnit:
		"""
		Logical units require logger and resource as init parameters, additionaly to the potentialy
		defined custom parameters which will be provided as a union of the model config
		and the kwargs provided to this method (the latter having prevalance)
		:raises: ValueError is the unit defined in the model is unknown
		"""
		return self.new(
			model,
			unit_type = sub_type or LogicalUnit,
			logger = logger,
			resource = self.get_resources(model),
			**kwargs
		)


	def new_safe_logical_unit(self,
		um: UnitModel,
		unit_type: type[LT],
		logger: AmpelLogger,
		_chan: None | ChannelId = None
	) -> LT:
		""" Returns a logical unit with dedicated logger containing no db handler """

		if logger.verbose:
			logger.log(VERBOSE, f"Instantiating unit {um.unit}")

		buf_hdlr = ChanRecordBufHandler(logger.level, _chan, {'unit': um.unit}) if _chan \
			else DefaultRecordBufferingHandler(logger.level, {'unit': um.unit})

		# Spawn unit instance
		unit = self.new_logical_unit(
			model = um,
			logger = AmpelLogger(
				base_flag = (getattr(logger, 'base_flag', 0) & ~LogFlag.CORE) | LogFlag.UNIT,
				console = len(logger.handlers) == 1, # to be improved later
				handlers = [buf_hdlr]
			),
			sub_type = unit_type
		)

		setattr(unit, '_buf_hdlr', buf_hdlr) # noqa: B010
		return unit


	@overload
	def new_context_unit(self,
		model: UnitModel, context: AmpelContext, *, sub_type: type[CT], **kwargs
	) -> CT:
		...
	@overload
	def new_context_unit(self,
		model: UnitModel, context: AmpelContext, *, sub_type: None = ..., **kwargs
	) -> ContextUnit:
		...
	def new_context_unit(self,
		model: UnitModel,
		context: AmpelContext, *,
		sub_type: None | type[CT] = None,
		**kwargs
	) -> CT | ContextUnit:
		"""
		Context units require an AmpelContext instance as init parameters, additionaly to
		potentialy defined dedicated custom parameters.
		:raises: ValueError is the unit defined in the model is unknown
		"""
		return self.new(
			model, unit_type=sub_type or ContextUnit, context=context, **kwargs
		)


	def get_adapter(self, context: AmpelContext, model: UnitModel, run_id: int) -> AbsUnitResultAdapter:
		akey = f"{context.uuid}_{run_id}_{build_unsafe_dict_id(model.dict())}"
		if (akey not in self._adapters):
			self._adapters[akey] = self.new(
				model, unit_type=AbsUnitResultAdapter, context=context, run_id = run_id
			)

		return self._adapters[akey]


	def clear_adapter_cache(self) -> None:
		self._adapters.clear()


	@overload
	def new(self, model: UnitModel, *, unit_type: type[T], **kwargs) -> T:
		...
	@overload
	def new(self, model: UnitModel, *, unit_type: None = ..., **kwargs) -> AmpelUnit:
		...
	def new(self, model: UnitModel, *, unit_type: None | type[T] = None, **kwargs) -> AmpelUnit | T:
		"""
		Instantiate new object based on provided model and kwargs.
		:param 'unit_type': performs isinstance check and raise error on mismatch. Enables mypy/other static checks.
		:returns: unit instance, trace id (0 if not computable)
		"""

		if not isinstance(model, UnitModel):
			raise ValueError(f"Unexpected model: '{type(model)}'")

		provenance = kwargs.pop('_provenance', self.provenance)

		Klass = self.get_class_by_name(model.unit, unit_type)
		if unit_type:
			check_class(Klass, unit_type)

		with NamedSecret.resolve_with(self.vault):
			unit = Klass(**(
				self.get_init_config(model.config, model.override)
				| kwargs
				| (model.secrets or {})
			))

			if hasattr(unit, "post_init"):
				unit.post_init()
		
		if isinstance(unit, LogicalUnit | ContextUnit):

			trace_id = None

			# potentially sync trace_ids with DB (Ampel_ext)
			if provenance:

				assert self.db

				trace_dict = {
					'py': pyv,
					'unit': model.unit,
					'digest': self.get_digest(Klass),
					'version': self.config.get(f"unit.{model.unit}.version", str, raise_exc=True)
				}

				if c := unit._get_trace_content():  # noqa: SLF001
					trace_dict['config'] = c

				if deps := self.config.get(f"unit.{model.unit}.dependencies"):
					if not isinstance(deps, list | tuple):
						raise ValueError(f"Retrieved environment is not a list/tuple: {type(deps)}")
					envd = self.config.get(f"environment.{env}", dict, raise_exc=True)
					trace_dict['env'] = {k: envd[k] for k in deps}

				try:

					# Note: we could implement a hash collision detection mechanism here
					trace_id = build_unsafe_dict_id(trace_dict, ret=int)

					# Save trace id to external collection
					if trace_id not in self.db.trace_ids:
						trace_dict['_id'] = trace_id
						self.db.add_trace_id(trace_id, trace_dict)

				# Non-serializable content
				except Exception:
					trace_id = 0
					# raise e

			unit._trace_id = trace_id  # noqa: SLF001

		return unit


	@staticmethod
	def get_digest(Klass: type) -> str:

		try:
			return blake2b(
				Path(sys.modules[Klass.__module__].__file__).read_bytes() # type: ignore[arg-type]
			).hexdigest()[:7]
		except Exception:
			return "unspecified"


	@overload
	def get_class_by_name(self, name: str, unit_type: type[T]) -> type[T]:
		...
	@overload
	def get_class_by_name(self, name: str, unit_type: None = ...) -> type[AmpelUnit]:
		...
	def get_class_by_name(self, name: str, unit_type: None | type[T] = None) -> type[T | AmpelUnit]:
		"""
		Matches the parameter 'name' with the unit definitions defined in the ampel_config.
		This allows to retrieve the corresponding fully qualified name of the class and to load it.

		:param unit_type:
			- LogicalUnit or any sublcass of LogicalUnit
			- ContextUnit or any sublcass of ContextUnit
			- If None (auxiliary class), returned object will have type[Any]

		:raises: ValueError if unit cannot be found or loaded or if parent class is unrecognized
		"""
		with contextlib.suppress(ValueError):
			return AuxUnitRegister.get_aux_class(name, sub_type=unit_type)

		if self._dyn_register and name in self._dyn_register:
			return self._dyn_register[name]

		if name in self.unit_defs:
			fqn = self.unit_defs[name]['fqn']
		else:
			raise ValueError(f"Ampel unit not found: {name}")

		# Note: importlib.import_module caches internally imported modules
		return getattr(import_module(fqn), name)


	def get_init_config(self,
		config: None | int | str | dict[str, Any] = None,
		override: None | dict[str, Any] = None,
		kwargs: None | dict[str, Any] = None,
		unfreeze: bool = True
	) -> dict[str, Any]:
		""" :raises: ValueError if config alias is not found """

		base_conf: dict[str, Any] = {}

		if isinstance(config, dict | str):

			base_conf = self.resolve_aliases(config)
			if base_conf is None:
				raise ValueError(f"Config alias {config} not found")

		# Hashed t2 unit configs
		elif isinstance(config, int):

			try:
				d = self.config.get_conf_by_id(config)
			# confid not found (obsolete or dynamically generated by isolated process)
			except Exception as e:
				assert self.db
				l = list(self.db.col_conf_ids.find({"_id": config}))
				if len(l) == 0:
					raise e
				del l[0]['_id']
				d = l[0]
				
			base_conf = recursive_unfreeze(d) if (unfreeze and isinstance(d, ReadOnlyDict)) else d

			# save un-registered (in ampel conf but not in db) confid to external collection for posterity
			if self.provenance:
				assert self.db
				if config not in self.db.conf_ids:
					self.db.add_conf_id(config, base_conf)

		return merge_dicts([base_conf, override, kwargs]) or {}


	def resolve_aliases(self, value):
		"""
		Recursively resolve aliases from config
		"""
		if isinstance(value, str):
			for adict in self.aliases:
				if value in adict:
					return self.resolve_aliases(adict[value])
			return value
		if isinstance(value, list):
			return [self.resolve_aliases(v) for v in value]
		if isinstance(value, dict):
			return {k: self.resolve_aliases(v) for k, v in value.items()}

		return value


	def get_resources(self, model: UnitModel) -> dict[str, Any]:
		"""
		Resources are defined using the static variable 'require' in ampel units
		-> example: catsHTM.default
		"""

		resources: dict[str, Any] = {}
		Klass = self.get_class_by_name(model.unit)

		# Load possibly required global resources
		for k in ampel_iter(getattr(Klass, 'require', [])):

			if k is None:
				continue

			# Global resource example: extcat
			if (resource := self.config.get(f'resource.{k}')) is None:
				raise ValueError(f"Global resource not available: {k}")

			resources[k] = resource

		return resources


	@contextmanager
	def validate_unit_models(self) -> Iterator[None]:
		""" Enable validation for UnitModel instances """

		@staticmethod # type: ignore[misc]
		def validate_unit(value: UnitModel) -> UnitModel:
			Unit = self.get_class_by_name(value.unit)
			if issubclass(Unit, AmpelUnit):
				Unit.validate(self.get_init_config(value.config, value.override))
			return value

		UnitModel.post_validate_hook = validate_unit
		AliasableModel._config = self.config  # noqa: SLF001
		try:
			with NamedSecret.resolve_with(self.vault):
				yield
		finally:
			UnitModel.post_validate_hook = None
			AliasableModel._config = None  # noqa: SLF001


	"""
	def internal_mypy_tests_uncomment_only_in_your_editor(self,
		model: UnitModel, context: AmpelContext, logger: AmpelLogger, sub_type: None | type[CT] = None, **kwargs
	) -> None:

		# Interal: uncomment to check if mypy works adequately
		from ampel.abstract.AbsEventUnit import AbsEventUnit
		from ampel.abstract.AbsLightCurveT2Unit import AbsLightCurveT2Unit

		reveal_type(self.new(model))
		reveal_type(self.new(model, bla=12))
		reveal_type(self.new(model, unit_type = None))
		reveal_type(self.new(model, unit_type=AbsLightCurveT2Unit))
		reveal_type(self.new(model, unit_type=AbsLightCurveT2Unit, bla=12))
		reveal_type(self.new(model, unit_type=AbsEventUnit))
		reveal_type(self.new(model, unit_type=AbsEventUnit, bla=12))

		reveal_type(self.new_logical_unit(model, logger))
		reveal_type(self.new_logical_unit(model, logger, bla=12))
		reveal_type(self.new_logical_unit(model, logger, sub_type = None))
		reveal_type(self.new_logical_unit(model, logger, sub_type=AbsLightCurveT2Unit))
		reveal_type(self.new_logical_unit(model, logger, sub_type = AbsLightCurveT2Unit, bla=12))

		# Next two lines *should* fail
		reveal_type(self.new_logical_unit(model, logger, sub_type=AbsEventUnit))
		reveal_type(self.new_logical_unit(model, logger, sub_type = AbsEventUnit, bla=12))

		reveal_type(self.new_context_unit(model, context))
		reveal_type(self.new_context_unit(model, context, bla=12))
		reveal_type(self.new_context_unit(model, context, sub_type = None))
		reveal_type(self.new_context_unit(model, context, sub_type = AbsEventUnit))
		reveal_type(self.new_context_unit(model, context, sub_type = AbsEventUnit, bla=12))

		# Next two lines *should* fail
		reveal_type(self.new_context_unit(model, context, sub_type = AbsLightCurveT2Unit))
		reveal_type(self.new_context_unit(model, context, sub_type = AbsLightCurveT2Unit, bla=12))
	"""



import datetime
import json
import subprocess
from os import environ
from pathlib import Path, PosixPath
from typing import Any

import mongomock
import mongomock.collection
import pymongo
import pytest

from ampel.abstract.AbsIngester import AbsIngester
from ampel.config.builder.DisplayOptions import DisplayOptions
from ampel.config.builder.DistConfigBuilder import DistConfigBuilder
from ampel.content.JournalRecord import JournalRecord
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.ingest.ChainedIngestionHandler import ChainedIngestionHandler
from ampel.ingest.StockCompiler import StockCompiler
from ampel.ingest.T2Compiler import T2Compiler
from ampel.log.AmpelLogger import DEBUG, AmpelLogger
from ampel.model.ingest.CompilerOptions import CompilerOptions
from ampel.model.ingest.IngestBody import IngestBody
from ampel.model.ingest.IngestDirective import IngestDirective
from ampel.model.ingest.T1Combine import T1Combine
from ampel.model.ingest.T2Compute import T2Compute
from ampel.model.UnitModel import UnitModel
from ampel.mongo.update.DBUpdatesBuffer import DBUpdatesBuffer
from ampel.mongo.update.MongoIngester import MongoIngester
from ampel.mongo.update.MongoStockIngester import MongoStockIngester
from ampel.mongo.update.MongoT2Ingester import MongoT2Ingester
from ampel.test.dummy import (
    DummyPointT2Unit,
    DummyStateT2Unit,
    DummyStockT2Unit,
    DummyTiedStateT2Unit,
)
from ampel.util.config import get_unit_confid


def pytest_addoption(parser):
    parser.addoption(
        "--integration",
        action="store_true",
        default=False,
        help="run docker-based integration tests",
    )


@pytest.fixture(scope="session")
def mongod(pytestconfig):
    if port := environ.get("MONGO_PORT"):
        yield f"mongodb://localhost:{port}"
        return

    if not pytestconfig.getoption("--integration"):
        raise pytest.skip("integration tests require --integration flag")
    try:
        container = subprocess.check_output(
            ["docker", "run", "--rm", "-d", "-P", "mongo:4.4"]
        ).decode().strip()
    except FileNotFoundError:
        pytest.skip("integration tests require docker")
    try:
        port = json.loads(subprocess.check_output(["docker", "inspect", container]))[0][
            "NetworkSettings"
        ]["Ports"]["27017/tcp"][0]["HostPort"]
        # wait for startup
        with pymongo.MongoClient(port=int(port)) as client:
            list(client.list_databases())
        yield f"mongodb://localhost:{port}"
    finally:
        ...
        subprocess.check_call(["docker", "stop", container])


@pytest.fixture
def _patch_mongo(monkeypatch):
    monkeypatch.setattr("ampel.core.AmpelDB.MongoClient", mongomock.MongoClient)
    # ignore codec_options in DataLoader
    monkeypatch.setattr("mongomock.codec_options.is_supported", lambda *args: None)
    # work around https://github.com/mongomock/mongomock/issues/912
    add_update = mongomock.collection.BulkOperationBuilder.add_update
    def _add_update(self, *args, sort=None, **kwargs):
        if sort is not None:
            raise NotImplementedError("sort not implemented in mongomock")
        return add_update(self, *args, **kwargs)
    monkeypatch.setattr("mongomock.collection.BulkOperationBuilder.add_update", _add_update)


@pytest.fixture(scope="session")
def testing_config():
    return Path(__file__).parent / "test-data" / "testing-config.yaml"


@pytest.fixture(scope="session")
def core_config():
    """
    The config distributed with ampel-core
    """
    cb = DistConfigBuilder(options=DisplayOptions())
    for ext in "json", "yml", "yaml":
        cb.load_distrib("ampel-core", "conf", ext)
    return cb.build_config(config_validator=None, get_unit_env=False)


@pytest.fixture
def mock_context(_patch_mongo, testing_config: PosixPath):
    return DevAmpelContext.load(config=str(testing_config), purge_db=True)


@pytest.fixture
def integration_context(mongod, testing_config: PosixPath):
    ctx = DevAmpelContext.load(
        config=str(testing_config),
        purge_db=True,
        custom_conf={"resource.mongo": mongod},
    )
    yield ctx
    ctx.db.close()


# metafixture as suggested in https://github.com/pytest-dev/pytest/issues/349#issuecomment-189370273
@pytest.fixture(params=["mock_context", "integration_context"])
def dev_context(request):
    return request.getfixturevalue(request.param)


@pytest.fixture
def ampel_logger():
    return AmpelLogger.get_logger(console=dict(level=DEBUG))


@pytest.fixture
def _ingest_stock(integration_context, ampel_logger):
    run_id = 0
    updates_buffer = DBUpdatesBuffer(integration_context.db, run_id=run_id, logger=ampel_logger)
    ingester = MongoStockIngester(
        updates_buffer=updates_buffer,
    )
    compiler = StockCompiler(
        run_id=run_id,
        tier=0,
    )
    compiler.add(
        "stockystock",
        channel="TEST_CHANNEL",
        journal=JournalRecord(tier=0, extra={"alert": 123}),
    )
    compiler.commit(ingester, datetime.datetime.now(tz=datetime.timezone.utc))
    ingester.updates_buffer.push_updates()
    assert integration_context.db.get_collection("stock").count_documents({}) == 1


@pytest.fixture
def _ingest_stock_t2(integration_context: DevAmpelContext, ampel_logger):
    run_id = 0
    updates_buffer = DBUpdatesBuffer(integration_context.db, run_id=run_id, logger=ampel_logger)
    stock_id = "stockystock"
    now = datetime.datetime.now(datetime.timezone.utc).timestamp()
    t2_compiler = T2Compiler(run_id=run_id, tier=2)
    t2_compiler.add(
        unit="DummyStockT2Unit",
        config=None,
        stock=stock_id,
        link=stock_id,
        channel="TEST_CHANNEL",
        ttl=None,
        traceid={},
    )
    t2_compiler.commit(
        MongoT2Ingester(updates_buffer=updates_buffer), now
    )
    
    stock_compiler = StockCompiler(run_id=run_id, tier=2)
    stock_compiler.add(stock_id, "TEST_CHANNEL", {"tier": 2})
    stock_compiler.commit(
        MongoStockIngester(updates_buffer=updates_buffer), now
    )
    
    updates_buffer.push_updates()
    assert integration_context.db.get_collection("stock").count_documents({}) == 1
    assert integration_context.db.get_collection("t2").count_documents({}) == 1

    integration_context.register_unit(DummyStockT2Unit)

def make_tied_ingestion_handler(
    integration_context: DevAmpelContext,
    logger: AmpelLogger,
    dependency: str,
    ingester_model=UnitModel(unit="MongoIngester")  # noqa: B008
) -> ChainedIngestionHandler:

    for unit in (
        DummyStockT2Unit,
        DummyPointT2Unit,
        DummyStateT2Unit,
        DummyTiedStateT2Unit,
    ):
        integration_context.register_unit(unit)

    unit_conf = {"t2_dependency": [{"unit": dependency}]}
    tied_config_id = get_unit_confid(integration_context.loader, unit="DummyTiedStateT2Unit", config=unit_conf)
    integration_context.add_conf_id(tied_config_id, unit_conf)
    run_id = 0
    channel = "TEST_CHANNEL"

    if "stock" in dependency.lower():
        body = IngestBody(
            stock_t2=[T2Compute(unit=dependency)],
            combine=[T1Combine(unit="T1SimpleCombiner", state_t2 = [T2Compute(unit="DummyTiedStateT2Unit", config=tied_config_id)])],
        )
    elif "point" in dependency.lower():
        body = IngestBody(
            point_t2=[T2Compute(unit=dependency)],
            combine=[
                T1Combine(unit="T1SimpleCombiner", state_t2=[T2Compute(unit="DummyTiedStateT2Unit", config=tied_config_id)])
            ],
        )
    else:
        body = IngestBody(
            combine=[
                T1Combine(unit="T1SimpleCombiner", state_t2=[T2Compute(unit=dependency), T2Compute(unit="DummyTiedStateT2Unit", config=tied_config_id)])
            ]
        )

    run_id = 0
    ingester = integration_context.loader.new_context_unit(
        ingester_model,
        context = integration_context,
        sub_type = AbsIngester,
        logger = logger,
        run_id = run_id,
        tier = 0,
        process_name = "ingest",
        raise_exc = True,
    )

    return ChainedIngestionHandler(
        integration_context,
        tier=0,
        run_id=run_id,
        trace_id={},
        ingester=ingester,
        logger=logger,
        compiler_opts=CompilerOptions(),
        shaper=UnitModel(unit="NoShaper"),
        directives=[IngestDirective(channel=channel, ingest=body)],
    )


@pytest.fixture(params=["DummyStateT2Unit", "DummyPointT2Unit", "DummyStockT2Unit"])
def ingest_tied_t2(integration_context: DevAmpelContext, ampel_logger, request):
    """Create a T2 document with dependencies"""

    for unit in (
        DummyStockT2Unit,
        DummyPointT2Unit,
        DummyStateT2Unit,
        DummyTiedStateT2Unit,
        MongoIngester,
    ):
        integration_context.register_unit(unit)
    
    handler = make_tied_ingestion_handler(integration_context, ampel_logger, request.param)

    datapoints: list[dict[str, Any]] = [
        {"id": i, "stock": "stockystock", "body": {"thing": i + 1}} for i in range(3)
    ]

    handler.ingest(datapoints, [(0, True)], stock_id="stockystock", jm_extra={"alert": 123})
    assert isinstance(handler.ingester, MongoIngester)

    handler.ingester._updates_buffer.push_updates()

    assert (
        integration_context.db.get_collection("t2").count_documents({"unit": request.param})
        == len(datapoints)
        if "point" in request.param.lower()
        else 1
    )
    assert (
        integration_context.db.get_collection("t2").count_documents(
            {"unit": "DummyTiedStateT2Unit"}
        )
        == 1
    )

    return request
